# -*- coding: utf-8 -*-

from itertools import chain
from bisect import bisect_left, bisect_right

import numpy as np
import scipy.sparse.linalg as splinalg

from .basis import BSplineBasis
from .splineobject import SplineObject
from .utils import ensure_listlike, is_singleton

__all__ = ['Curve']


class Curve(SplineObject):
    """Curve()

    Represents a curve: an object with a one-dimensional parameter space."""

    _intended_pardim = 1

    def __init__(self, basis=None, controlpoints=None, rational=False, **kwargs):
        """  Construct a curve with the given basis and control points.

        The default is to create a linear one-element mapping from (0,1) to the
        unit interval.

        :param BSplineBasis basis: The underlying B-Spline basis
        :param array-like controlpoints: An *n* × *d* matrix of control points
        :param bool rational: Whether the curve is rational (in which case the
            control points are interpreted as pre-multiplied with the weight,
            which is the last coordinate)
        """
        super(Curve, self).__init__([basis], controlpoints, rational, **kwargs)

    def evaluate(self, *params):
        """  Evaluate the object at given parametric values.

        This function returns an *n1* × *n2* × ... × *dim* array, where *ni* is
        the number of evaluation points in direction *i*, and *dim* is the
        physical dimension of the object.

        If there is only one evaluation point, a vector of length *dim* is
        returned instead.

        :param u,v,...: Parametric coordinates in which to evaluate
        :type u,v,...: float or [float]
        :return: Geometry coordinates
        :rtype: numpy.array
        """
        squeeze = is_singleton(params[0])
        params = [ensure_listlike(p) for p in params]

        self._validate_domain(*params)

        # Evaluate the derivatives of the corresponding bases at the corresponding points
        # and build the result array
        N = self.bases[0].evaluate(params[0], sparse=True)
        result = N @ self.controlpoints

        # For rational objects, we divide out the weights, which are stored in the
        # last coordinate
        if self.rational:
            for i in range(self.dimension):
                result[..., i] /= result[..., -1]
            result = np.delete(result, self.dimension, -1)

        # Squeeze the singleton dimensions if we only have one point
        if squeeze:
            result = result.reshape(self.dimension)

        return result

    def derivative(self, t, d=1, above=True, tensor=True):
        """  Evaluate the derivative of the curve at the given parametric values.

        This function returns an *n* × *dim* array, where *n* is the number of
        evaluation points, and *dim* is the physical dimension of the curve.

        If there is only one evaluation point, a vector of length *dim* is
        returned instead.

        :param t: Parametric coordinates in which to evaluate
        :type t: float or [float]
        :param int d: Number of derivatives to compute
        :param bool above: Evaluation in the limit from above
        :param bool tensor: Not used in this method
        :return: Derivative array
        :rtype: numpy.array
        """
        if not is_singleton(d):
            d = d[0]
        if not self.rational or d < 2 or d > 3:
            return super(Curve, self).derivative(t, d=d, above=above, tensor=tensor)

        t = ensure_listlike(t)
        result = np.zeros((len(t), self.dimension))

        d2 = np.array(self.bases[0].evaluate(t, 2, above) @ self.controlpoints)
        d1 = np.array(self.bases[0].evaluate(t, 1, above) @ self.controlpoints)
        d0 = np.array(self.bases[0].evaluate(t) @ self.controlpoints)
        W  = d0[:, -1]  # W(t)
        W1 = d1[:, -1]  # W'(t)
        W2 = d2[:, -1]  # W''(t)
        if d == 2:
            for i in range(self.dimension):
                result[:, i] = (d2[:, i] * W * W - 2 * W1 *
                               (d1[:, i] * W - d0[:, i] * W1) - d0[:, i] * W2 * W) / W / W / W
        if d == 3:
            d3 = np.array(self.bases[0].evaluate(t, 3, above) @ self.controlpoints)
            W3 = d3[:,-1]    # W'''(t)
            W6 = W*W*W*W*W*W # W^6
            for i in range(self.dimension):
                H  =  d1[:,i]*W                           - d0[:,i]*W1
                H1 =  d2[:,i]*W                           - d0[:,i]*W2
                H2 =  d3[:,i]*W + d2[:,i]*W1 - d1[:,i]*W2 - d0[:,i]*W3
                G  =  H1*W - 2*H*W1
                G1 =  H2*W - 2*H*W2 - H1*W1
                result[:, i] = (G1*W - 3*G*W1) /W/W/W/W

        if result.shape[0] == 1:  # in case of single value input t, return vector instead of matrix
            result = np.array(result[0, :]).reshape(self.dimension)

        return result

    def binormal(self, t, above=True):
        """  Evaluate the normalized binormal of the curve at the given parametric value(s).

        This function returns an *n* × 3 array, where *n* is the number of
        evaluation points.

        The binormal is computed as the normalized cross product between the
        velocity and acceleration of the curve.

        :param t: Parametric coordinates in which to evaluate
        :type t: float or [float]
        :param bool above: Evaluation in the limit from above
        :return: Derivative array
        :rtype: numpy.array
        """
        # error test input
        if self.dimension != 3:
            raise ValueError('Binormals require dimension = 3')

        # compute derivative
        dx    = self.derivative(t, d=1, above=above)
        ddx   = self.derivative(t, d=2, above=above)

        # in case of vanishing acceleration, colinear velocity and acceleration,
        # such as linear curves we guess an appropriate binbormal (multiple choice available)
        if len(dx.shape) == 1:
            if np.allclose(ddx, 0):
                if np.allclose(dx[:2], 0): # dx = [0,0,1]
                    ddx = np.array([1,0,0])
                else:
                    ddx = np.array([0,0,1])
        else:
            for i in range(ddx.shape[0]):
                if np.allclose(ddx[i,:], 0):
                    if np.allclose(dx[i,:2], 0): # dx = [0,0,1]
                        ddx[i,:] = np.array([1,0,0])
                    else:
                        ddx[i,:] = np.array([0,0,1])

        result = np.cross(dx, ddx)

        # in case of single value input t, return vector instead of matrix
        if len(dx.shape) == 1:
            return result / np.linalg.norm(result)

        # normalize
        magnitude = np.linalg.norm(result, axis=1)
        magnitude = magnitude.reshape(-1,1)

        return result / magnitude

    def normal(self, t, above=True):
        """  Evaluate the normal of the curve at the given parametric value(s).

        This function returns an *n* × 3 array, where *n* is the number of
        evaluation points.

        The normal is computed as the cross product between the binormal and
        the tangent of the curve.

        :param t: Parametric coordinates in which to evaluate
        :type t: float or [float]
        :param bool above: Evaluation in the limit from above
        :return: Derivative array
        :rtype: numpy.array
        """
        # error test input
        if self.dimension != 3:
            raise RuntimeError('Normals require dimension = 3')

        # compute derivative
        T = self.tangent(t,  above=above)
        B = self.binormal(t, above=above)

        return np.cross(B,T)

    def curvature(self, t, above=True):
        """  Evaluate the curvaure at specified point(s). The curvature is defined as

        .. math:: \\frac{|\\boldsymbol{v}\\times \\boldsymbol{a}|}{|\\boldsymbol{v}|^3}

        :param t: Parametric coordinates in which to evaluate
        :type t: float or [float]
        :param bool above: Evaluation in the limit from above
        :return: Derivative array
        :rtype: numpy.array
        """
        # compute derivative
        v = self.derivative(t, d=1, above=above)
        a = self.derivative(t, d=2, above=above)
        w = np.cross(v,a)

        if len(v.shape) == 1: # single evaluation point
            magnitude = np.linalg.norm(w)
            speed     = np.linalg.norm(v)
        else:                 # multiple evaluation points
            if self.dimension == 2:
                # for 2D-cases np.cross() outputs scalars
                # (the z-component of the cross product)
                magnitude = np.abs(w)
            else:
                # for 3D, it is vectors
                magnitude = np.linalg.norm( w,  axis= -1)

            speed = np.linalg.norm( v, axis=-1)

        return magnitude / np.power(speed,3)

    def torsion(self, t, above=True):
        """  Evaluate the torsion for a 3D curve at specified point(s). The torsion is defined as

        .. math:: \\frac{(\\boldsymbol{v}\\times \\boldsymbol{a})\\cdot (d\\boldsymbol{a}/dt)}{|\\boldsymbol{v}\\times \\boldsymbol{a}|^2}

        :param t: Parametric coordinates in which to evaluate
        :type t: float or [float]
        :param bool above: Evaluation in the limit from above
        :return: Derivative array
        :rtype: numpy.array
        """
        if self.dimension == 2:
            # no torsion for 2D curves
            t = ensure_listlike(t)
            return np.zeros(len(t))
        elif self.dimension == 3:
            # only allow 3D curves
            pass
        else:
            raise ValueError('dimension must be 2 or 3')

        # compute derivative
        v  = self.derivative(t, d=1, above=above)
        a  = self.derivative(t, d=2, above=above)
        da = self.derivative(t, d=3, above=above)
        w = np.cross(v,a)

        if len(v.shape) == 1: # single evaluation point
            magnitude = np.linalg.norm(w)
            nominator = np.dot(w, a)
        else:                 # multiple evaluation points
            magnitude = np.linalg.norm( w, axis=-1)
            nominator = np.array([np.dot(w1,da1) for (w1,da1) in zip(w, da)])

        return nominator / np.power(magnitude, 2)

    def raise_order(self, amount, direction=None):
        """  Raise the polynomial order of the curve.

        :param int amount: Number of times to raise the order
        :return: self
        """
        if amount < 0:
            raise ValueError('Raise order requires a non-negative parameter')
        elif amount == 0:
            return

        # create the new basis
        newBasis = self.bases[0].raise_order(amount)

        # set up an interpolation problem. This is in projective space, so no problems for rational cases
        interpolation_pts_t = newBasis.greville()  # parametric interpolation points (t)
        N_old = self.bases[0].evaluate(interpolation_pts_t)
        N_new = newBasis.evaluate(interpolation_pts_t, sparse=True)
        interpolation_pts_x = N_old @ self.controlpoints  # projective interpolation points (x,y,z,w)

        # solve the interpolation problem
        self.controlpoints = np.array(splinalg.spsolve(N_new, interpolation_pts_x))
        self.bases = [newBasis]

        return self

    def append(self, curve):
        """  Extend the curve by merging another curve to the end of it.

        The curves are glued together in a C0 fashion with enough repeated
        knots. The function assumes that the end of this curve perfectly
        matches the start of the input curve.

        :param Curve curve: Another curve
        :raises RuntimeError: If either curve is periodic
        :return: self
        """
        # ASSUMPTION: open knot vectors

        # error test input
        if self.bases[0].periodic > -1 or curve.bases[0].periodic > -1:
            raise RuntimeError('Cannot append with periodic curves')

        # copy input curve so we don't change that one directly
        extending_curve = curve.clone()

        # make sure both are in the same space, and (if needed) have rational weights
        Curve.make_splines_compatible(self, extending_curve)

        # make sure both have the same discretization order
        p1 = self.order(0)
        p2 = extending_curve.order(0)
        if p1 < p2:
            self.raise_order(p2 - p1)
        else:
            extending_curve.raise_order(p1 - p2)
        p = max(p1, p2)

        # build new knot vector by merging the two existing ones
        old_knot = self.knots(direction=0, with_multiplicities=True)
        add_knot = extending_curve.knots(direction=0, with_multiplicities=True)
        # make sure that the new one starts where the old one stops
        add_knot -= add_knot[0]
        add_knot += old_knot[-1]
        new_knot = np.zeros(len(add_knot) + len(old_knot) - p - 1)
        new_knot[:len(old_knot) - 1] = old_knot[:-1]
        new_knot[len(old_knot) - 1:] = add_knot[p:]

        # build new control points by merging the two existing matrices
        n1 = len(self)
        n2 = len(extending_curve)
        new_controlpoints = np.zeros((n1 + n2 - 1, self.dimension + self.rational))
        new_controlpoints[:n1, :] = self.controlpoints
        new_controlpoints[n1:, :] = extending_curve.controlpoints[1:, :]

        # update basis and controlpoints
        self.bases = [BSplineBasis(p, new_knot)]
        self.controlpoints = new_controlpoints

        return self

    def continuity(self, knot):
        """  Get the parametric continuity of the curve at a given point. Will
        return p-1-m, where m is the knot multiplicity and inf between knots"""
        return self.bases[0].continuity(knot)

    def get_kinks(self):
        """  Get the parametric coordinates at all points which have C0-
        continuity"""
        return [k for k in self.knots(0) if self.continuity(k)<1]

    def length(self, t0=None, t1=None):
        """ Computes the euclidian length of the curve in geometric space

        .. math:: \\int_{t_0}^{t_1}\\sqrt{x(t)^2 + y(t)^2 + z(t)^2} dt

        """
        (x,w) = np.polynomial.legendre.leggauss(self.order(0)+1)
        knots = self.knots(0)
        # keep only integration boundaries within given start (t0) and stop (t1) interval
        if t0 is not None:
            i = bisect_left(knots, t0)
            knots = np.insert(knots, i, t0)
            knots = knots[i:]
        if t1 is not None:
            i = bisect_right(knots, t1)
            knots = knots[:i]
            knots = np.insert(knots, i, t1)

        t = np.array([ (x+1)/2*(t1-t0)+t0 for t0,t1 in zip(knots[:-1], knots[1:]) ])
        w = np.array([     w/2*(t1-t0)    for t0,t1 in zip(knots[:-1], knots[1:]) ])
        t = np.ndarray.flatten(t)
        w = np.ndarray.flatten(w)
        dx = self.derivative(t)
        detJ = np.sqrt(np.sum(dx**2, axis=1))
        return np.dot(detJ, w)

    def rebuild(self, p, n):
        """  Creates an approximation to this curve by resampling it using a
        uniform knot vector of order *p* with *n* control points.

        :param int p: Polynomial discretization order
        :param int n: Number of control points
        :return: A new approximate curve
        :rtype: Curve
        """
        # establish uniform open knot vector
        knot = [0] * p + list(range(1, n - p + 1)) + [n - p + 1] * p
        basis = BSplineBasis(p, knot)
        # set parametric range of the new basis to be the same as the old one
        basis.normalize()
        t0 = self.bases[0].start()
        t1 = self.bases[0].end()
        basis *= (t1 - t0)
        basis += t0
        # fetch evaluation points and solve interpolation problem
        t = basis.greville()
        N = basis.evaluate(t, sparse=True)
        controlpoints = splinalg.spsolve(N, self.evaluate(t))

        # return new resampled curve
        return Curve(basis, controlpoints)

    def error(self, target):
        """  Computes the L2 (squared and per knot span) and max error between
        this curve and a target curve

        .. math:: ||\\boldsymbol{x_h}(t)-\\boldsymbol{x}(t)||_{L^2(t_1,t_2)}^2 = \\int_{t_1}^{t_2}
            |\\boldsymbol{x_h}(t)-\\boldsymbol{x}(t)|^2 dt, \\quad \\forall \\;\\text{knots}\\;t_1 < t_2

        .. math:: ||\\boldsymbol{x_h}(t)-\\boldsymbol{x}(t)||_{L^\\infty} = \\max_t |\\boldsymbol{x_h}(t)-\\boldsymbol{x}(t)|

        :param function target: callable function which takes as input a vector
            of evaluation points t and gives as output a matrix x where
            x[i,j] is component j evaluated at point t[i]
        :return: L2 error per knot-span and the maximum error
        :rtype:  tuple(list(float), float)

        Examples:

        .. code:: python

            import numpy as np

            def arclength_circle(t):
                return np.array( [np.cos(t), np.sin(t)] ).T

            crv = curve_factory.circle(r=1)
            (err2, maxerr) = crv.error(arclength_circle)
            print('|| e ||_L2  = ', np.sqrt(np.sum(err2)))
            print('|| e ||_max = ', maxerr)
        """
        knots = self.knots(0)
        (x,w) = np.polynomial.legendre.leggauss(self.order(0)+1)
        err2    = []
        err_inf = 0.0
        for t0,t1 in zip(knots[:-1], knots[1:]): # for all knot spans
            tg  = (x+1)/2*(t1-t0)+t0 # evaluation points
            wg  =   w  /2*(t1-t0)    # integration weights
            error = self(tg) - target(tg)    # [x-xh, y-yh, z-zh]
            error = np.sum(error**2, axis=1) # |x-xh|^2
            err2.append(np.dot(error, wg))   # integrate over domain
            err_inf = max(np.max(np.sqrt(error)), err_inf)
        return (err2, err_inf)

    def __repr__(self):
        return str(self.bases[0]) + '\n' + str(self.controlpoints)

    def get_derivative_curve(self):
        return super(Curve, self).get_derivative_spline(0)



# -*- coding: utf-8 -*-

from itertools import chain

import numpy as np

from .basis import BSplineBasis
from .splineobject import SplineObject
from .utils import ensure_listlike, check_direction, sections

__all__ = ['Volume']


class Volume(SplineObject):
    """Volume()

    Represents a volume: an object with a three-dimensional parameter space."""

    _intended_pardim = 3

    def __init__(self, basis1=None, basis2=None, basis3=None, controlpoints=None, rational=False, **kwargs):
        """  Construct a volume with the given basis and control points.

        The default is to create a linear one-element mapping from and to the
        unit cube.

        :param BSplineBasis basis1: The basis of the first parameter direction
        :param BSplineBasis basis2: The basis of the second parameter direction
        :param BSplineBasis basis3: The basis of the third parameter direction
        :param array-like controlpoints: An *n1* × *n2* × *n3* × *d* matrix of
            control points
        :param bool rational: Whether the volume is rational (in which case the
            control points are interpreted as pre-multiplied with the weight,
            which is the last coordinate)
        """
        super(Volume, self).__init__([basis1, basis2, basis3], controlpoints, rational, **kwargs)

    def edges(self):
        """Return the twelve edges of this volume in order:

        - umin, vmin
        - umax, vmin
        - umin, vmax
        - umax, vmax
        - umin, wmin
        - umax, wmin
        - umin, wmax
        - umax, wmax
        - vmin, wmin
        - vmax, wmin
        - vmin, wmax
        - vmax, wmax

        :return: Edges
        :rtype: (Curve)
        """
        return tuple(self.section(*args) for args in sections(3, 1))

    def faces(self):
        """Return the six faces of this volume in order: umin, umax, vmin, vmax, wmin, wmax.

        :return: Boundary faces
        :rtype: (Surface)
        """
        return tuple(self.section(*args) for args in sections(3, 2))

    def volume(self):
        """ Computes the volume of the object in geometric space """
        # fetch integration points
        (x1,w1) = np.polynomial.legendre.leggauss(self.order(0)+1)
        (x2,w2) = np.polynomial.legendre.leggauss(self.order(1)+1)
        (x3,w3) = np.polynomial.legendre.leggauss(self.order(2)+1)
        # map points to parametric coordinates (and update the weights)
        (knots1,knots2,knots3) = self.knots()
        u  = np.array([ (x1+1)/2*(t1-t0)+t0 for t0,t1 in zip(knots1[:-1], knots1[1:]) ])
        w1 = np.array([     w1/2*(t1-t0)    for t0,t1 in zip(knots1[:-1], knots1[1:]) ])
        v  = np.array([ (x2+1)/2*(t1-t0)+t0 for t0,t1 in zip(knots2[:-1], knots2[1:]) ])
        w2 = np.array([     w2/2*(t1-t0)    for t0,t1 in zip(knots2[:-1], knots2[1:]) ])
        w  = np.array([ (x3+1)/2*(t1-t0)+t0 for t0,t1 in zip(knots3[:-1], knots3[1:]) ])
        w3 = np.array([     w3/2*(t1-t0)    for t0,t1 in zip(knots3[:-1], knots3[1:]) ])

        # wrap everything to vectors
        u = np.ndarray.flatten(u)
        v = np.ndarray.flatten(v)
        w = np.ndarray.flatten(w)
        w1 = np.ndarray.flatten(w1)
        w2 = np.ndarray.flatten(w2)
        w3 = np.ndarray.flatten(w3)

        # compute all quantities of interest (i.e. the jacobian)
        du = self.derivative(u,v,w, d=(1,0,0))
        dv = self.derivative(u,v,w, d=(0,1,0))
        dw = self.derivative(u,v,w, d=(0,0,1))

        J  = du[:,:,:,0] * np.cross(dv[:,:,:,1:],   dw[:,:,:,1:]  ) -  \
             du[:,:,:,1] * np.cross(dv[:,:,:,0::2], dw[:,:,:,0::2]) +  \
             du[:,:,:,2] * np.cross(dv[:,:,:,:-1],  dw[:,:,:,:-1] )

        return np.abs(J).dot(w3).dot(w2).dot(w1)

    def rebuild(self, p, n):
        """  Creates an approximation to this volume by resampling it using
        uniform knot vectors of order *p* with *n* control points.

        :param (int) p: Tuple of polynomial discretization order in each direction
        :param (int) n: Tuple of number of control points in each direction
        :return: A new approximate volume
        :rtype: Volume
        """
        p = ensure_listlike(p, dups=3)
        n = ensure_listlike(n, dups=3)

        old_basis = [self.bases[0], self.bases[1], self.bases[2]]
        basis = []
        u = []
        N = []
        # establish uniform open knot vectors
        for i in range(3):
            knot = [0] * p[i] + list(range(1, n[i] - p[i] + 1)) + [n[i] - p[i] + 1] * p[i]
            basis.append(BSplineBasis(p[i], knot))

            # make these span the same parametric domain as the old ones
            basis[i].normalize()
            t0 = old_basis[i].start()
            t1 = old_basis[i].end()
            basis[i] *= (t1 - t0)
            basis[i] += t0

            # fetch evaluation points and evaluate basis functions
            u.append(basis[i].greville())
            N.append(basis[i].evaluate(u[i]))

        # find interpolation points as evaluation of existing volume
        x = self.evaluate(u[0], u[1], u[2])

        # solve interpolation problem
        cp = np.tensordot(np.linalg.inv(N[2]), x, axes=(1, 2))
        cp = np.tensordot(np.linalg.inv(N[1]), cp, axes=(1, 2))
        cp = np.tensordot(np.linalg.inv(N[0]), cp, axes=(1, 2))

        # re-order controlpoints so they match up with Volume constructor
        cp = cp.transpose((2, 1, 0, 3))
        cp = cp.reshape(n[0] * n[1] * n[2], cp.shape[3])

        # return new resampled curve
        return Volume(basis[0], basis[1], basis[2], cp)

    def __repr__(self):
        result = str(self.bases[0]) + '\n'
        result += str(self.bases[1]) + '\n'
        result += str(self.bases[2]) + '\n'
        # print legacy controlpoint enumeration
        n1, n2, n3, dim = self.controlpoints.shape
        for k in range(n3):
            for j in range(n2):
                for i in range(n1):
                    result += str(self.controlpoints[i, j, k, :]) + '\n'
        return result

    get_derivative_volume = SplineObject.get_derivative_spline



from itertools import groupby
from operator import itemgetter
from os.path import exists, isdir, join
from os import makedirs

import numpy as np

from ..splinemodel import SplineModel


class OpenFOAM(object):

    def __init__(self, target):
        self.target = target

    def __enter__(self):
        # Create the target directory if it does not exist
        if not exists(self.target):
            makedirs(target)
        # If it does, ensure that it's a directory
        elif not isdir(self.target):
            raise FileExistsError('{} exists and is not a directory'.format(self.target))

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def _header(self, cls, obj, note=None):
        s = 'FoamFile\n{\n'
        s += '    version     2.0;\n'
        s += '    format      ascii;\n'
        s += '    class       %s;\n' % cls
        if note:
            s += '    note        "%s";\n' % note
        s += '    object      %s;\n' % obj
        s += '}\n'
        return s

    def write(self, model):
        assert isinstance(model, SplineModel), "OpenFOAM.write only supports SplineModel objects"

        # Only linear volumes in 3D, please
        assert model.pardim == 3
        assert model.dimension == 3
        assert all(patch.obj.order() == (2,2,2) for patch in model.catalogue.top_nodes())

        # Generate all the information we need
        model.generate_cp_numbers()
        model.generate_cell_numbers()
        faces = model.faces()
        ninternal = sum(faces['name'] == None)
        note = 'nPoints: {} nCells: {} nFaces: {} nInternalFaces: {}'.format(
            model.ncps, model.ncells, len(faces), ninternal
        )

        # OpenFOAM is very particular about face ordering
        # In order of importance:
        # - Internal faces before boundary faces
        # - All faces in the same boundary must be contiguous
        # - Low number owners before high number owners
        # - Low number neighbors before high number neighbors
        faces = list(faces)
        faces = sorted(faces, key=itemgetter('neighbor'))
        faces = sorted(faces, key=itemgetter('owner'))
        faces = sorted(faces, key=lambda x: (x['name'] is not None, x['name']))
        faces = np.array(faces)

        # Write the points file (vertex coordinates)
        with open(join(self.target, 'points'), 'w') as f:
            f.write(self._header('vectorField', 'points'))
            f.write(str(model.ncps) + '\n(\n')
            for pt in model.cps():
                f.write('({})\n'.format(' '.join(str(p) for p in pt)))
            f.write(')\n')

        # Write the faces file (four vertex indices for each face)
        with open(join(self.target, 'faces'), 'w') as f:
            f.write(self._header('faceList', 'faces'))
            f.write(str(len(faces)) + '\n(\n')
            for face in faces['nodes']:
                f.write('({})\n'.format(' '.join(str(f) for f in face)))
            f.write(')\n')

        # Write the owner and neighbour files (cell indices for each face)
        with open(join(self.target, 'owner'), 'w') as f:
            f.write(self._header('labelList', 'owner', note=note))
            f.write(str(len(faces)) + '\n(\n')
            for owner in faces['owner']:
                f.write(str(owner) + '\n')
            f.write(')\n')
        with open(join(self.target, 'neighbour'), 'w') as f:
            f.write(self._header('labelList', 'neighbour', note=note))
            f.write(str(len(faces)) + '\n(\n')
            for neighbor in faces['neighbor']:
                f.write(str(neighbor) + '\n')
            f.write(')\n')

        # Write the boundary file
        with open(join(self.target, 'boundary'), 'w') as f:
            f.write(self._header('polyBoundaryMesh', 'boundary'))
            f.write(str(len(set(faces['name'])) - 1) + '\n(\n')
            start = 0
            for name, it in groupby(faces, key=itemgetter('name')):
                nfaces = len(list(it))
                if name is None:
                    start += nfaces
                    continue
                f.write(name + '\n{\n')
                f.write('    type patch;\n')
                f.write('    nFaces {};\n'.format(nfaces))
                f.write('    startFace {};\n'.format(start))
                f.write('}\n')
                start += nfaces
            f.write(')\n')



from .basis import BSplineBasis
from .splineobject import SplineObject
from .curve import Curve
from .surface import Surface
from .volume import Volume
from .trimmedsurface import TrimmedSurface
from .splinemodel import SplineModel

__version__ = '1.5.8'



# -*- coding: utf-8 -*-

"""Handy utilities for creating volumes."""

from math import pi, sqrt, atan2

import numpy as np

from .basis import BSplineBasis
from .surface import Surface
from .volume import Volume
from .utils import flip_and_move_plane_geometry, rotate_local_x_axis
from . import curve_factory, surface_factory

__all__ = ['cube', 'sphere', 'revolve', 'cylinder', 'extrude', 'edge_surfaces',
           'loft', 'interpolate', 'least_square_fit']



def cube(size=1, lower_left=(0,0,0)):
    """  Create a cube with parmetric origin at *(0,0,0)*.

    :param float size: Size(s), either a single scalar or a tuple of scalars per axis
    :param array-like lower_left: local origin, the lower bottom left corner of the cube
    :return: A linear parametrized box
    :rtype: Volume
    """
    result = Volume()
    result.scale(size)
    result += lower_left
    return result

def sphere(r=1, center=(0,0,0), type='radial'):
    """  Create a solid sphere

    :param float r: Radius
    :param array-like center: Local origin of the sphere
    :param string type: The type of parametrization ('radial' or 'square')
    :return: A solid ball
    :rtype: Volume
    """
    if type == 'radial':
        shell    = surface_factory.sphere(r, center)
        midpoint = shell*0 + center
        return edge_surfaces(shell, midpoint)
    elif type == 'square':
        # based on the work of James E.Cobb: "Tiling the Sphere with Rational Bezier Patches"
        # University of Utah, July 11, 1988. UUCS-88-009
        b = BSplineBasis(order=5)
        sr2 = sqrt(2)
        sr3 = sqrt(3)
        sr6 = sqrt(6)
        cp = [[      -4*(sr3-1),       4*(1-sr3),      4*(1-sr3),   4*(3-sr3)  ], # row 0
              [           -sr2 ,     sr2*(sr3-4),    sr2*(sr3-4), sr2*(3*sr3-2)],
              [              0 ,  4./3*(1-2*sr3), 4./3*(1-2*sr3),4./3*(5-sr3)  ],
              [            sr2 ,     sr2*(sr3-4),    sr2*(sr3-4), sr2*(3*sr3-2)],
              [       4*(sr3-1),       4*(1-sr3),      4*(1-sr3),   4*(3-sr3)  ],
              [    -sr2*(4-sr3),            -sr2,    sr2*(sr3-4), sr2*(3*sr3-2)], # row 1
              [    -(3*sr3-2)/2,     (2-3*sr3)/2,     -(sr3+6)/2,     (sr3+6)/2],
              [              0 , sr2*(2*sr3-7)/3,       -5*sr6/3, sr2*(sr3+6)/3],
              [     (3*sr3-2)/2,     (2-3*sr3)/2,     -(sr3+6)/2,     (sr3+6)/2],
              [     sr2*(4-sr3),            -sr2,    sr2*(sr3-4), sr2*(3*sr3-2)],
              [ -4./3*(2*sr3-1),               0, 4./3*(1-2*sr3),   4*(5-sr3)/3], # row 2
              [-sr2/3*(7-2*sr3),               0,       -5*sr6/3, sr2*(sr3+6)/3],
              [              0 ,               0,    4*(sr3-5)/3, 4*(5*sr3-1)/9],
              [ sr2/3*(7-2*sr3),               0,       -5*sr6/3, sr2*(sr3+6)/3],
              [  4./3*(2*sr3-1),               0, 4./3*(1-2*sr3),   4*(5-sr3)/3],
              [    -sr2*(4-sr3),             sr2,    sr2*(sr3-4), sr2*(3*sr3-2)], # row 3
              [    -(3*sr3-2)/2,    -(2-3*sr3)/2,     -(sr3+6)/2,     (sr3+6)/2],
              [              0 ,-sr2*(2*sr3-7)/3,       -5*sr6/3, sr2*(sr3+6)/3],
              [     (3*sr3-2)/2,    -(2-3*sr3)/2,     -(sr3+6)/2,     (sr3+6)/2],
              [     sr2*(4-sr3),             sr2,    sr2*(sr3-4), sr2*(3*sr3-2)],
              [      -4*(sr3-1),      -4*(1-sr3),      4*(1-sr3),   4*(3-sr3)  ], # row 4
              [           -sr2 ,    -sr2*(sr3-4),    sr2*(sr3-4), sr2*(3*sr3-2)],
              [              0 , -4./3*(1-2*sr3), 4./3*(1-2*sr3),4./3*(5-sr3)  ],
              [            sr2 ,    -sr2*(sr3-4),    sr2*(sr3-4), sr2*(3*sr3-2)],
              [       4*(sr3-1),      -4*(1-sr3),      4*(1-sr3),   4*(3-sr3)  ]]
        wmin = Surface(b,b,cp, rational=True)
        wmax = wmin.clone().mirror([0,0,1])
        vmax = wmin.clone().rotate(pi/2, [1,0,0])
        vmin = vmax.clone().mirror([0,1,0])
        umax = vmin.clone().rotate(pi/2, [0,0,1])
        umin = umax.clone().mirror([1,0,0])
        # ideally I would like to call edge_surfaces() now, but that function
        # does not work with rational surfaces, so we'll just manually try
        # and add some inner controlpoints
        cp   = np.zeros((5,5,5,4))
        cp[ :, :, 0,:] = wmin[:,:,:]
        cp[ :, :,-1,:] = wmax[:,:,:]
        cp[ :, 0, :,:] = vmin[:,:,:]
        cp[ :,-1, :,:] = vmax[:,:,:]
        cp[ 0, :, :,:] = umin[:,:,:]
        cp[-1, :, :,:] = umax[:,:,:]
        inner = np.linspace(-.5,.5, 3)
        Y, X, Z = np.meshgrid(inner,inner,inner)
        cp[1:4,1:4,1:4,0] = X
        cp[1:4,1:4,1:4,1] = Y
        cp[1:4,1:4,1:4,2] = Z
        cp[1:4,1:4,1:4,3] = 1
        ball = Volume(b,b,b,cp,rational=True, raw=True)
        return r*ball + center
    else:
        raise ValueError('invalid type argument')


def revolve(surf, theta=2 * pi, axis=(0,0,1)):
    """  Revolve a volume by sweeping a surface in a rotational fashion around
    an axis.

    :param Surface surf: Surface to revolve
    :param float theta: Angle to revolve, in radians
    :param array-like axis: Axis of rotation
    :return: The revolved surface
    :rtype: Volume
    """
    surf = surf.clone()  # clone input surface, throw away old reference
    surf.set_dimension(3)  # add z-components (if not already present)
    surf.force_rational()  # add weight (if not already present)

    # align axis with the z-axis
    normal_theta = atan2(axis[1], axis[0])
    normal_phi   = atan2(sqrt(axis[0]**2 + axis[1]**2), axis[2])
    surf.rotate(-normal_theta, [0,0,1])
    surf.rotate(-normal_phi,   [0,1,0])

    path = curve_factory.circle_segment(theta=theta)
    n = len(surf)  # number of control points of the surface
    m = len(path)  # number of control points of the sweep

    cp = np.zeros((m * n, 4))

    dt = np.sign(theta)*(path.knots(0)[1] - path.knots(0)[0]) / 2.0
    for i in range(m):
        weight = path[i,-1]
        cp[i * n:(i + 1) * n, :] = np.reshape(surf.controlpoints.transpose(1, 0, 2), (n, 4))
        cp[i * n:(i + 1) * n, 2] *= weight
        cp[i * n:(i + 1) * n, 3] *= weight
        surf.rotate(dt)
    result = Volume(surf.bases[0], surf.bases[1], path.bases[0], cp, True)
    # rotate it back again
    result.rotate(normal_phi,   [0,1,0])
    result.rotate(normal_theta, [0,0,1])
    return result

def torus(minor_r=1, major_r=3, center=(0,0,0), normal=(0,0,1), xaxis=(1,0,0), type='radial'):
    """  Create a torus (doughnut) by revolving a circle of size *minor_r*
    around the *z* axis with radius *major_r*.

    :param float minor_r: The thickness of the torus (radius in the *xz* plane)
    :param float major_r: The size of the torus (radius in the *xy* plane)
    :param array-like center: Local origin of the torus
    :param array-like normal: Local origin of the torus
    :param array-like center: Local origin of the torus
    :param string type: The type of parametrization ('radial' or 'square')
    :return: A periodic torus
    :rtype: Volume
    """

    disc = surface_factory.disc(minor_r, type=type)
    disc.rotate(pi / 2, (1, 0, 0))  # flip up into xz-plane
    disc.translate((major_r, 0, 0))  # move into position to spin around z-axis
    result = revolve(disc)

    result.rotate(rotate_local_x_axis(xaxis, normal))
    return flip_and_move_plane_geometry(result, center, normal)

def cylinder(r=1, h=1, center=(0,0,0), axis=(0,0,1), xaxis=(1,0,0), type='radial'):
    """  Create a solid cylinder

    :param float r: Radius
    :param float h: Height
    :param array-like center: The center of the bottom circle
    :param array-like axis: Cylinder axis
    :param array-like xaxis: direction of sem, i.e. parametric start point u=0
    :param string type: The type of parametrization ('radial' or 'square')
    :return: The cylinder
    :rtype: Volume
    """
    return extrude(surface_factory.disc(r, center, axis, xaxis=xaxis, type=type), h*np.array(axis))


def extrude(surf, amount):
    """  Extrude a surface by sweeping it to a given height.

    :param Surface surf: Surface to extrude
    :param array-like amount: 3-component vector of sweeping amount and direction
    :return: The extruded surface
    :rtype: Volume
    """
    surf.set_dimension(3)  # add z-components (if not already present)
    cp = []
    for controlpoint in surf:
        cp.append(list(controlpoint))
    surf += amount
    for controlpoint in surf:
        cp.append(list(controlpoint))
    surf -= amount
    return Volume(surf.bases[0], surf.bases[1], BSplineBasis(2), cp, surf.rational)


def edge_surfaces(*surfaces):
    """  Create the volume defined by the region between the input surfaces.

    In case of six input surfaces, these must be given in the order: bottom,
    top, left, right, back, front. Opposing sides must be parametrized in the
    same directions.

    :param [Surface] surfaces: Two or six edge surfaces
    :return: The enclosed volume
    :rtype: Volume
    :raises ValueError: If the length of *surfaces* is not two or six
    """
    if len(surfaces) == 1: # probably gives input as a list-like single variable
        surfaces = surfaces[0]
    if len(surfaces) == 2:
        surf1 = surfaces[0].clone()
        surf2 = surfaces[1].clone()
        Surface.make_splines_identical(surf1, surf2)
        (n1, n2, d) = surf1.controlpoints.shape  # d = dimension + rational

        controlpoints = np.zeros((n1, n2, 2, d))
        controlpoints[:, :, 0, :] = surf1.controlpoints
        controlpoints[:, :, 1, :] = surf2.controlpoints

        # Volume constructor orders control points in a different way, so we
        # create it from scratch here
        result = Volume(surf1.bases[0], surf1.bases[1], BSplineBasis(2), controlpoints,
                         rational=surf1.rational, raw=True)

        return result
    elif len(surfaces) == 6:
        if any([surf.rational for surf in surfaces]):
            raise RuntimeError('edge_surfaces not supported for rational splines')

        # coons patch (https://en.wikipedia.org/wiki/Coons_patch)
        umin = surfaces[0]
        umax = surfaces[1]
        vmin = surfaces[2]
        vmax = surfaces[3]
        wmin = surfaces[4]
        wmax = surfaces[5]
        vol1 = edge_surfaces(umin,umax)
        vol2 = edge_surfaces(vmin,vmax)
        vol3 = edge_surfaces(wmin,wmax)
        vol4 = Volume(controlpoints=vol1.corners(order='F'), rational=vol1.rational)
        vol1.swap(0, 2)
        vol1.swap(1, 2)
        vol2.swap(1, 2)
        vol4.swap(1, 2)
        Volume.make_splines_identical(vol1, vol2)
        Volume.make_splines_identical(vol1, vol3)
        Volume.make_splines_identical(vol1, vol4)
        Volume.make_splines_identical(vol2, vol3)
        Volume.make_splines_identical(vol2, vol4)
        Volume.make_splines_identical(vol3, vol4)
        result                =   vol1.clone()
        result.controlpoints +=   vol2.controlpoints
        result.controlpoints +=   vol3.controlpoints


		### as suggested by github user UnaiSan (see issues 141)
        result.controlpoints += vol4.controlpoints

        Nu, Nv, Nw, d = result.controlpoints.shape

        controlpoints = np.zeros((2, 2, Nw, d))
        controlpoints[0, 0, :] = vol1.controlpoints[0, 0]
        controlpoints[0, 1, :] = vol1.controlpoints[0, -1]
        controlpoints[1, 0, :] = vol1.controlpoints[-1, 0]
        controlpoints[1, 1, :] = vol1.controlpoints[-1, -1]
        vol_u_edges = Volume(
            BSplineBasis(),
            BSplineBasis(),
            result.bases[2],
            controlpoints=controlpoints,
            raw=True,
            rational=result.rational
        )

        controlpoints = np.zeros((Nu, 2, 2, d))
        controlpoints[:, 0, 0] = vol2.controlpoints[:, 0, 0]
        controlpoints[:, 0, 1] = vol2.controlpoints[:, 0, -1]
        controlpoints[:, 1, 0] = vol2.controlpoints[:, -1, 0]
        controlpoints[:, 1, 1] = vol2.controlpoints[:, -1, -1]
        vol_v_edges = Volume(
            result.bases[0],
            BSplineBasis(),
            BSplineBasis(),
            controlpoints=controlpoints,
            raw=True,
            rational=result.rational
        )

        controlpoints = np.zeros((2, Nv, 2, d))
        controlpoints[0, :, 0] = vol3.controlpoints[0, :, 0]
        controlpoints[0, :, 1] = vol3.controlpoints[0, :, -1]
        controlpoints[1, :, 0] = vol3.controlpoints[-1, :, 0]
        controlpoints[1, :, 1] = vol3.controlpoints[-1, :, -1]
        vol_w_edges = Volume(
            BSplineBasis(),
            result.bases[1],
            BSplineBasis(),
            controlpoints=controlpoints,
            raw=True,
            rational=result.rational
        )

        Volume.make_splines_identical(result, vol_u_edges)
        Volume.make_splines_identical(result, vol_v_edges)
        Volume.make_splines_identical(result, vol_w_edges)

        result.controlpoints -= vol_u_edges.controlpoints
        result.controlpoints -= vol_v_edges.controlpoints
        result.controlpoints -= vol_w_edges.controlpoints

        return result
    else:
        raise ValueError('Requires two or six input surfaces')

def sweep(path, shape):
    """  Generate a surface by sweeping a shape along a path

    The resulting surface is an approximation generated by interpolating at the
    Greville points. It is generated by sweeping a shape curve along a path.

    The *shape* object has to be contained in the 'xy' plane (preferably centered
    around the origin) as its x-coordinate is extruded in the normal direction,
    and its y-coordinate in the binormal direction of the *path* curve.

    :param Curve path:  The path to drag *shape* along
    :param Surface shape: The shape to be dragged out to a surface
    :return: Surrounding volume
    :rtype: Volume
    """
    b1 = shape.bases[0]
    b2 = shape.bases[1]
    b3 = path.bases[0]
    n1 = b1.num_functions()
    n2 = b2.num_functions()
    n3 = b3.num_functions()
    # this requires binormals and normals, which only work in 3D, so assume this here
    X  = np.zeros((n1,n2,n3, 3))

    # pre-evaluate the surface
    u = b1.greville()
    v = b2.greville()
    y = shape(u,v)

    for k in range(n3):
        w = b3.greville(k)
        x = path(w)
        B = path.binormal(w)
        N = path.normal(w)
        for i in range(n1):
            for j in range(n2):
                X[i,j,k,:] = x + N*y[i,j,0] + B*y[i,j,1]

    return interpolate(X, [b1,b2,b3])


def loft(*surfaces):
    """  Generate a volume by lofting a series of surfaces

    The resulting volume is interpolated at all input surfaces and a smooth transition
    between these surfaces is computed as a cubic spline interpolation in the lofting
    direction. In the case that insufficient surfaces are provided as input (less than 4
    surfaces), then a quadratic or linear interpolation is performed.

    Note that the order of input surfaces matter as they will be interpolated in this
    particular order. Also note that the surfaces need to be parametrized in the same
    direction, otherwise you will encounter self-intersecting result volume as it is
    wrapping in on itself.

    :param Surfaces surfaces:  A sequence of surfaces to be lofted
    :return: Lofted volume
    :rtype: Volume

    Examples:

    .. code:: python

        from splipy import surface_factory, volume_factory
        srf1 = surface_factory.disc(r=1)
        srf2 = 1.5 * srf1 + [0,0,1]
        srf3 = 1.0 * srf1 + [0,0,2]
        srf4 = 1.3 * srf1 + [0,0,4]
        vol = volume_factory.loft(srf1, srf2, srf3, srf4)

        # alternatively you can provide all input curves as a list
        all_my_surfaces = [srf1, srf2, srf3, srf4]
        vol = volume_factory.loft(all_my_surfaces)

    """
    if len(surfaces) == 1:
        surfaces = surfaces[0]

    # clone input, so we don't change those references
    # make sure everything has the same dimension since we need to compute length
    surfaces = [s.clone().set_dimension(3) for s in surfaces]
    if len(surfaces)==2:
        return surface_factory.edge_curves(surfaces)
    elif len(surfaces)==3:
        # can't do cubic spline interpolation, so we'll do quadratic
        basis3 = BSplineBasis(3)
        dist  = basis3.greville()
    else:
        x = [s.center() for s in surfaces]

        # create knot vector from the euclidian length between the surfaces
        dist = [0]
        for (x1,x0) in zip(x[1:],x[:-1]):
            dist.append(dist[-1] + np.linalg.norm(x1-x0))

        # using "free" boundary condition by setting N'''(u) continuous at second to last and second knot
        knot = [dist[0]]*4 + dist[2:-2] + [dist[-1]]*4
        basis3 = BSplineBasis(4, knot)

    n = len(surfaces)
    for i in range(n):
        for j in range(i+1,n):
            Surface.make_splines_identical(surfaces[i], surfaces[j])

    basis1 = surfaces[0].bases[0]
    basis2 = surfaces[0].bases[1]
    m1     = basis1.num_functions()
    m2     = basis2.num_functions()
    dim    = len(surfaces[0][0])
    u      = basis1.greville() # parametric interpolation points
    v      = basis2.greville()
    w      = dist

    # compute matrices
    Nu     = basis1(u)
    Nv     = basis2(v)
    Nw     = basis3(w)
    Nu_inv = np.linalg.inv(Nu)
    Nv_inv = np.linalg.inv(Nv)
    Nw_inv = np.linalg.inv(Nw)

    # compute interpolation points in physical space
    x      = np.zeros((m1,m2,n, dim))
    for i in range(n):
        tmp        = np.tensordot(Nv, surfaces[i].controlpoints, axes=(1,1))
        x[:,:,i,:] = np.tensordot(Nu, tmp                      , axes=(1,1))

    # solve interpolation problem
    cp = np.tensordot(Nw_inv, x,  axes=(1,2))
    cp = np.tensordot(Nv_inv, cp, axes=(1,2))
    cp = np.tensordot(Nu_inv, cp, axes=(1,2))

    # re-order controlpoints so they match up with Surface constructor
    cp = np.reshape(cp.transpose((2, 1, 0, 3)), (m1*m2*n, dim))

    return Volume(basis1, basis2, basis3, cp, surfaces[0].rational)


def interpolate(x, bases, u=None):
    """  Interpolate a volume on a set of regular gridded interpolation points `x`.

    The points can be either a matrix (in which case the first index is
    interpreted as a flat row-first index of the interpolation grid) or a 4D
    tensor. In both cases the last index is the physical coordinates.

    :param numpy.ndarray x: Grid of interpolation points
    :param [BSplineBasis] bases: The basis to interpolate on
    :param [array-like] u: Parametric interpolation points, defaults to
        Greville points of the basis
    :return: Interpolated volume
    :rtype: Volume
    """
    vol_shape = [b.num_functions() for b in bases]
    dim = x.shape[-1]
    if len(x.shape) == 2:
        x = x.reshape(vol_shape + [dim])
    if u is None:
        u = [b.greville() for b in bases]
    N_all = [b(t) for b,t in zip(bases, u)]
    N_all.reverse()
    cp = x
    for N in N_all:
        cp = np.tensordot(np.linalg.inv(N), cp, axes=(1,2))

    return Volume(bases[0], bases[1], bases[2], cp.transpose(2,1,0,3).reshape((np.prod(vol_shape),dim)))


def least_square_fit(x, bases, u):
    """  Perform a least-square fit of a point cloud `x` onto a spline basis.

    The points can be either a matrix (in which case the first index is
    interpreted as a flat row-first index of the interpolation grid) or a 4D
    tensor. In both cases the last index is the physical coordinates.

    There must be at least as many points as basis functions.

    :param numpy.ndarray x: Grid of evaluation points
    :param [BSplineBasis] bases: Basis on which to interpolate
    :param [array-like] u: Parametric values at evaluation points
    :return: Approximated volume
    :rtype: Volume
    """
    vol_shape = [b.num_functions() for b in bases]
    dim = x.shape[-1]
    if len(x.shape) == 2:
        x = x.reshape(vol_shape + [dim])
    N_all = [b(t) for b,t in zip(bases, u)]
    N_all.reverse()
    cp = x
    for N in N_all:
        cp = np.tensordot(N.T, cp, axes=(1,2))
    for N in N_all:
        cp = np.tensordot(np.linalg.inv(N.T @ N), cp, axes=(1,2))

    return Volume(bases[0], bases[1], bases[2], cp.transpose(2,1,0,3).reshape((np.prod(vol_shape),dim)))



import numpy as np
from itertools import chain, product
from splipy import Curve, Surface, BSplineBasis, curve_factory
from .master import MasterIO
import splipy.state as state
from rhino3dm import Brep, File3dm
from rhino3dm import NurbsCurve, PolylineCurve, Circle, Polyline, BezierCurve, Arc, Line
from rhino3dm import NurbsSurface, Cylinder, Sphere, Extrusion
from rhino3dm import Curve   as threedmCurve   # name conflict with splipy
from rhino3dm import Surface as threedmSurface # name conflict with splipy

class ThreeDM(MasterIO):

    def __init__(self, filename):
        if filename[-4:] != '.3dm':
            filename += '.3dm'
        self.filename = filename
        self.trimming_curves = []

    def __enter__(self):
        return self

    def write(self, obj):
        raise IOError('Writing to 3DM not supported')

    def read(self):
        if not hasattr(self, 'fstream'):
            self.onlywrite = False
            self.fstream   = File3dm.Read(self.filename)

        if self.onlywrite:
            raise IOError('Could not read from file %s' % (self.filename))

        result = []

        for obj in self.fstream.Objects:
            geom = obj.Geometry
            print(geom)
            if type(geom) is Extrusion:
                geom = geom.ToBrep(splitKinkyFaces=True)
            if type(geom) is Brep:
                for idx in range(len(geom.Faces)):
                    print('  ', geom.Faces[idx], "(", geom.Faces[idx].UnderlyingSurface(), ")")
                    nsrf = geom.Faces[idx].UnderlyingSurface().ToNurbsSurface()
                    result.append(self.read_surface(nsrf))

            if type(geom) is Line:
                geom = result.append(curve_factory.line(geom.From, geom.To))
                continue
            if type(geom) is PolylineCurve:
                geom = geom.ToPolyline()
            if type(geom) is Polyline or \
               type(geom) is Circle or \
               type(geom) is threedmCurve or \
               type(geom) is BezierCurve or \
               type(geom) is Arc:
                geom = geom.ToNurbsCurve()

            if type(geom) is NurbsCurve:
                result.append(self.read_curve(geom))

            if type(geom) is Cylinder or \
               type(geom) is Sphere or \
               type(geom) is threedmSurface:
                geom = geom.ToNurbsSurface()
            if type(geom) is NurbsSurface:
                result.append(self.read_surface(geom))

        return result

    def read_surface(self, nsrf):
        knotsu = [0]
        for i in nsrf.KnotsU:
            knotsu.append(i)
        knotsu.append(knotsu[len(knotsu)-1])
        knotsu[0] = knotsu[1]

        knotsv = [0]
        for i in nsrf.KnotsV:
            knotsv.append(i)
        knotsv.append(knotsv[len(knotsv)-1])
        knotsv[0] = knotsv[1]

        basisu = BSplineBasis(nsrf.OrderU, knotsu, -1)
        basisv = BSplineBasis(nsrf.OrderV, knotsv, -1)
        cpts = []

        cpts = np.ndarray((nsrf.Points.CountU*nsrf.Points.CountV, 3 + nsrf.IsRational))
        for v in range(0,nsrf.Points.CountV):
            for u in range(0,nsrf.Points.CountU):
                cpts[u+v*nsrf.Points.CountU,0] = nsrf.Points[u,v].X
                cpts[u+v*nsrf.Points.CountU,1] = nsrf.Points[u,v].Y
                cpts[u+v*nsrf.Points.CountU,2] = nsrf.Points[u,v].Z
                if nsrf.IsRational:
                    cpts[u+v*nsrf.Points.CountU,3] = nsrf.Points[u,v].W

        return Surface(basisu, basisv, cpts, nsrf.IsRational)

    def read_curve(self, ncrv):
        knots = [0]
        for i in ncrv.Knots:
            knots.append(i)
        knots[0] = knots[1]
        knots.append(knots[len(knots)-1])
        basis = BSplineBasis(ncrv.Order, knots, -1)
        cpts = []

        cpts = np.ndarray((len(ncrv.Points), ncrv.Dimension + ncrv.IsRational))
        for u in range(0,len(ncrv.Points)):
            cpts[u,0] = ncrv.Points[u].X
            cpts[u,1] = ncrv.Points[u].Y
            if ncrv.Dimension > 2:
                cpts[u,2] = ncrv.Points[u].Z
            if ncrv.IsRational:
                cpts[u,3] = ncrv.Points[u].W

        return Curve(basis, cpts, ncrv.IsRational)

    def __exit__(self, exc_type, exc_value, traceback):
        # Apperently File3DM objects don't need to dedicated cleanup/close code
        pass



# -*- coding: utf-8 -*-

import numpy as np

from ..curve import Curve
from ..basis import BSplineBasis
from .. import surface_factory

__all__ = ['camber', 'NACA']


def camber(M, P, order=5):
    """ Create the NACA centerline used for wing profiles. This is given as
    an exact quadratic piecewise polynomial y(x),
    see http://airfoiltools.com/airfoil/naca4digit. The method will produce
    one of two representations: For order<5 it will create x(t)=t and
    for order>4 it will create x(t) as qudratic in t and stretched towards the
    endpointspoints creating a more optimized parametrization.
    :param M: Max camber height (y) given as percentage 0% to 9% of length
    :type  M: Int 0<M<10
    :param P: Max camber position (x) given as percentage 0% to 90% of length
    :type  P: Int 0<P<10
    :return : Exact centerline representation
    :rtype  : Curve
    """
    # parametrized by x=t or x="t^2" if order>4
    M = M / 100.0
    P = P / 10.0
    basis = BSplineBasis(order)
    # basis.insert_knot([P]*(order-2)) # insert a C1-knot
    for i in range(order - 2):
        basis.insert_knot(P)

    t = basis.greville()
    n = len(t)
    x = np.zeros((n, 2))
    for i in range(n):
        if t[i] <= P:
            if order > 4:
                x[i, 0] = t[i]**2 / P
            else:
                x[i, 0] = t[i]
            x[i, 1] = M / P / P * (2 * P * x[i, 0] - x[i, 0] * x[i, 0])
        else:
            if order > 4:
                x[i, 0] = (t[i]**2 - 2 * t[i] + P) / (P - 1)
            else:
                x[i, 0] = t[i]
            x[i, 1] = M / (1 - P) / (1 - P) * (1 - 2 * P + 2 * P * x[i, 0] - x[i, 0] * x[i, 0])
    N = basis.evaluate(t)
    controlpoints = np.linalg.solve(N, x)
    return Curve(basis, controlpoints)


def NACA(M, P, X, n=40, order=5, closed=False):
    """ Create the NACA 4 digit airfoil. This is generated as an approximation
    through the use of SurfaceFactory.thicken functions.
    :param M: Max camber height (y) given as percentage 0% to 9% of length
    :type  M: Int 0<M<10
    :param P: Max camber position (x) given as percentage 0% to 90% of length
    :type  P: Int 0<P<10
    :param X: Thickness given in percentage 1
    :type  X: Int 0<X<40
    :param n: Number of knot spans in resulting parametrization
    :type  n: Int
    :param order: Order of resulting spline curve
    :type  order: Int
    :return : Approximation of NACA wing profile
    :rtype  : Curve
    """

    n = int(n / 4.0)  # knot spans on each side
    center_line = camber(M, P, order)
    new_knots1 = np.linspace(0, P / 10.0, n + 2)
    new_knots2 = np.linspace(P / 10.0, 1, n + 2)
    new_knots = np.zeros(2 * n)
    new_knots[:n] = new_knots1[1:-1]
    new_knots[n:] = new_knots2[1:-1]
    center_line.insert_knot(new_knots)
    T = X / 100.0

    def thickness(x):
        a0 = 0.2969
        a1 = -0.126
        a2 = -0.3516
        a3 = 0.2843
        a4 = -0.1036 if closed else -0.1015
        return T / 0.2 * (a0 * np.sqrt(x) + a1 * x + a2 * x**2 + a3 * x**3 + a4 * x**4)

    surf = surface_factory.thicken(center_line, thickness)
    _, _, top, btm = surf.edges()
    top.reverse()
    top.append(btm)

    if closed:
        top = top.make_periodic(0)

    return top



# -*- coding: utf-8 -*-

from bisect import bisect_right, bisect_left
import copy

import numpy as np
from scipy.sparse import csr_matrix

from .utils import ensure_listlike
from . import basis_eval, state

__all__ = ['BSplineBasis']


class BSplineBasis:
    """BSplineBasis()

    Represents a one-dimensional B-Spline basis.

    BSplineBasis objects support basic arithmetic operators, which are
    interpreted as acting on the parametric domain.
    """
    knots = [0, 0, 1, 1]
    order = 2
    periodic = -1

    def __init__(self, order=2, knots=None, periodic=-1):
        """  Construct a B-Spline basis with a given order and knot vector.

        :param int order: Spline order, i.e. one greater than the polynomial degree.
        :param [float] knots: Knot vector of non-decreasing components.
            Defaults to open knot vector on domain [0,1].
        :param int periodic: Number of continuous derivatives at start and end.
            --1 is not periodic, 0 is continuous, etc.
        :raises ValueError: for inapproriate knot vectors
        """

        periodic = max(periodic, -1)
        if knots is None:
            knots = [0] * order + [1] * order
            for i in range(periodic+1):
                knots[   i] = -1
                knots[-i-1] =  2

        self.knots = np.array(knots)
        self.knots = self.knots.astype(float)
        self.order = order
        self.periodic = periodic

        # error test input
        p          = order
        k          = periodic
        n          = len(knots)
        if p < 1:
            raise ValueError('invalid spline order')
        if n < 2*p:
            raise ValueError('knot vector has too few elements')
        if periodic >= 0:
            for i in range(p + k - 1):
                if abs((knots[i + 1] - knots[i]) - (knots[-p - k + i ] - knots[-p - k - 1 + i])) > state.knot_tolerance:
                    raise ValueError('periodic knot vector is mis-matching at the start/end')
        for i in range(len(knots) - 1):
            if knots[i + 1] - knots[i] < -state.knot_tolerance:
                raise ValueError('knot vector needs to be non-decreasing')

    def num_functions(self):
        """  Returns the number of basis functions in the basis.

        .. warning:: This is different from :func:`splipy.BSplineBasis.__len__`."""
        return len(self.knots) - self.order - (self.periodic + 1)

    def start(self):
        """Start point of parametric domain. For open knot vectors, this is the
        first knot.

        :return: Knot number *p*, where *p* is the spline order
        :rtype: float
        """
        return self.knots[self.order - 1]

    def end(self):
        """End point of parametric domain. For open knot vectors, this is the
        last knot.

        :return: Knot number *n*--*p*, where *p* is the spline order and *n* is
            the number of knots
        :rtype: Float
        """
        return self.knots[-self.order]

    def greville(self, index=None):
        """  Fetch greville points, also known as knot averages:

        .. math:: \\sum_{j=i+1}^{i+p-1} \\frac{t_j}{p-1}

        :return: One, or all of the Greville points
        :rtype: [float] (if *index* is ``None``) or float
        """
        result = []
        p = self.order
        n = self.num_functions()
        if index is None:
            for i in range(n):
                result.append(float(np.sum(self.knots[i + 1:i + p])) / (p - 1))
        else:
            result = float(np.sum(self.knots[index + 1:index + p])) / (p - 1)
        return result

    def evaluate(self, t, d=0, from_right=True, sparse=False):
        """  Evaluate all basis functions in a given set of points.

        :param t: The parametric coordinate(s) in which to evaluate
        :type t: float or [float]
        :param int d: Number of derivatives to compute
        :param bool from_right: True if evaluation should be done in the limit
            from above
        :param bool sparse: True if computed matrix should be returned as sparse
        :return: A matrix *N[i,j]* of all basis functions *j* evaluated in all
            points *i*
        :rtype: numpy.array
        """

        # for single-value input, wrap it into a list so it don't crash on the loop below
        t = ensure_listlike(t)
        t = np.array(t, dtype=np.float64)
        basis_eval.snap(self.knots, t, state.knot_tolerance)

        if self.order <= d: # requesting more derivatives than polymoial degree: return all zeros
            return np.zeros((len(t), self.num_functions()))

        (data, size) = basis_eval.evaluate(self.knots, self.order, t, self.periodic, state.knot_tolerance, d, from_right)

        N = csr_matrix(data, size)
        if not sparse:
            N = N.toarray()
        return N

    def evaluate_old(self, t, d=0, from_right=True, sparse=False):
        """  Evaluate all basis functions in a given set of points.
        :param t: The parametric coordinate(s) in which to evaluate
        :type t: float or [float]
        :param int d: Number of derivatives to compute
        :param bool from_right: True if evaluation should be done in the limit
            from above
        :param bool sparse: True if computed matrix should be returned as sparse
        :return: A matrix *N[i,j]* of all basis functions *j* evaluated in all
            points *i*
        :rtype: numpy.array
        """
        # for single-value input, wrap it into a list so it don't crash on the loop below
        t = ensure_listlike(t)
        self.snap(t)

        p = self.order  # knot vector order
        n_all = len(self.knots) - p  # number of basis functions (without periodicity)
        n = len(self.knots) - p - (self.periodic+1)  # number of basis functions (with periodicity)
        m = len(t)
        data    = np.zeros(m*p)
        indices = np.zeros(m*p, dtype='int32')
        indptr  = np.array(range(0,m*p+1,p), dtype='int32')
        if p <= d: # requesting more derivatives than polymoial degree: return all zeros
            return np.zeros((m,n))
        if self.periodic >= 0:
            t = copy.deepcopy(t)
            # Wrap periodic evaluation into domain
            for i in range(len(t)):
                if t[i] < self.start() or t[i] > self.end():
                    t[i] = (t[i] - self.start()) % (self.end() - self.start()) + self.start()
        for i in range(len(t)):
            right = from_right
            evalT = t[i]
            # Special-case the endpoint, so the user doesn't need to
            if abs(t[i] - self.end()) < state.knot_tolerance:
                right = False
            # Skip non-periodic evaluation points outside the domain
            if t[i] < self.start() or t[i] > self.end():
                continue

            # mu = index of last non-zero basis function
            if right:
                mu = bisect_right(self.knots, evalT)
            else:
                mu = bisect_left(self.knots, evalT)
            mu = min(mu, n_all)

            M = np.zeros(p)  # temp storage to keep all the function evaluations
            M[-1] = 1  # the last entry is a dummy-zero which is never used
            for q in range(1, p-d):
                for j in range(p - q - 1, p):
                    k = mu - p + j  # 'i'-index in global knot vector (ref Hughes book pg.21)
                    if j != p-q-1:
                        M[j] = M[j] * float(evalT - self.knots[k]) / (self.knots[k + q] - self.knots[k])

                    if j != p-1:
                        M[j] = M[j] + M[j + 1] * float(self.knots[k + q + 1] - evalT) / (self.knots[k + q + 1] - self.knots[k + 1])


            for q in range(p-d, p):
                for j in range(p - q - 1, p):
                    k = mu - p + j  # 'i'-index in global knot vector (ref Hughes book pg.21)
                    if j != p-q-1:
                        M[j] = M[j] * float(q) / (self.knots[k + q] - self.knots[k])
                    if j != p-1:
                        M[j] = M[j] - M[j + 1] * float(q) / (self.knots[k + q + 1] - self.knots[k + 1])


            data[i*p:(i+1)*p]    = M
            indices[i*p:(i+1)*p] = np.arange(mu-p, mu) % n

        N = csr_matrix((data, indices, indptr), (m,n))
        if not sparse:
            N = N.toarray()
        return N

    def integrate(self, t0, t1):
        """  Integrate all basis functions over a given domain

        :param float t0: The parametric starting point
        :param float t1: The parametric end point
        :return: The integration of all functions over the input domain
        :rtype: list
        """
        if self.periodic > -1 and (t0<self.start() or t1>self.end()):
            raise NotImplemented('Periodic functions integrated across sem')

        t0 = max(t0, self.start())
        t1 = min(t1, self.end()  )
        p  = self.order
        knot = [self.knots[0]] + list(self.knots) + [self.knots[-1]]
        integration_basis = BSplineBasis(p + 1, knot)
        N0 = np.array(integration_basis.evaluate(t0)).flatten()
        N1 = np.array(integration_basis.evaluate(t1)).flatten()
        N  = [(knot[i+p]-knot[i])*1.0/p * np.sum(N1[i:]-N0[i:]) for i in range(N0.size)]
        N  = N[1:]

        # collapse periodic functions onto themselves
        if self.periodic > -1:
            for j in range(self.periodic + 1):
                N[j] += N[-self.periodic - 1 + j]
            N = N[:-self.periodic-1]

        return N

    def normalize(self):
        """Set the parametric domain to be (0,1)."""
        self -= self.start()  # set start-point to 0
        self /= self.end()  # set end-point to 1

    def reparam(self, start=0, end=1):
        """  Set the parametric domain to be (start, end)

        :raises ValueError: If *end* ≤ *start*"""
        if end <= start:
            raise ValueError('end must be larger than start')
        self.normalize()
        self *= (end - start)
        self += start

    def reverse(self):
        """Reverse parametric domain, keeping start/end values unchanged."""
        a = float(self.start())
        b = float(self.end())
        self.knots = (self.knots[::-1] - a) / (b - a) * (a - b) + b

    def continuity(self, knot):
        """Get the continuity of the basis functions at a given point.

        :return: *p*--*m*--1 at a knot with multiplicity *m*, or ``inf``
            between knots.
        :rtype: int or float
        """
        if self.periodic >= 0:
            if knot < self.start() or knot > self.end():
                knot = (knot - self.start()) % (self.end() - self.start()) + self.start()
        elif knot < self.start() or self.end() < knot:
            raise ValueError('out of range')

        # First knot that is larger than the right tolerance point
        hi = bisect_left(self.knots, knot + state.knot_tolerance)

        # Last knot that is smaller than the left tolerance point
        lo = bisect_left(self.knots, knot - state.knot_tolerance)

        if hi == lo:
            return np.inf
        return self.order - (hi - lo) - 1

    def make_periodic(self, continuity):
        """Create a periodic basis with a given continuity."""
        deg = self.order - 1
        new_knots = self.knots[deg:-deg]

        diff = self.end() - self.start()
        n_reps = deg - continuity - 1
        n_copy = deg - n_reps

        head = new_knots[-n_copy-1:-1] - diff
        tail = new_knots[1:n_copy+1] + diff

        new_knots = np.hstack((head, [self.start()] * n_reps, new_knots, [self.end()] * n_reps, tail))
        return BSplineBasis(self.order, new_knots, continuity)

    def knot_spans(self, include_ghost_knots=False):
        """Return the set of unique knots in the knot vector.

        :param bool include_ghost_knots: if knots outside start/end are to be
            included. These knots are used by periodic basis.
        :return: List of unique knots
        :rtype: [float]"""
        p = self.order
        if include_ghost_knots:
            result = [self.knots[0]]
            for k in self.knots:
                if abs(k - result[-1]) > state.knot_tolerance:
                    result.append(k)
        else:
            result = [self.knots[p-1]]
            for k in self.knots[p-1:-p+1]:
                if abs(k - result[-1]) > state.knot_tolerance:
                    result.append(k)
        return result

    def raise_order(self, amount):
        """Create a knot vector with higher order.

        The continuity at the knots are kept unchanged by increasing their
        multiplicities.

        :return: New knot vector
        :rtype: [float]
        :raises TypeError: If `amount` is not an int
        :raises ValueError: If `amount` is negative
        """
        if type(amount) is not int:
            raise TypeError('amount needs to be a non-negative integer')
        if amount < 0:
            raise ValueError('amount needs to be a non-negative integer')
        if amount == 0:
            return self.clone()
        knot_spans = list(self.knot_spans(True))  # list of unique knots
        # For every degree we raise, we need to increase the multiplicity by one
        knots = list(self.knots) + knot_spans * amount
        # make it a proper knot vector by ensuring that it is non-decreasing
        knots.sort()
        if self.periodic > -1:
            # remove excessive ghost knots which appear at both ends of the knot vector
            n0 =                   bisect_left(knot_spans, self.start())
            n1 = len(knot_spans) - bisect_left(knot_spans, self.end())   - 1
            knots = knots[n0*amount : -n1*amount]

        return BSplineBasis(self.order + amount, knots, self.periodic)

    def lower_order(self, amount):
        """Create a knot vector with lower order.

        The continuity at the knots are kept unchanged by decreasing their
        multiplicities.

        :return: New knot vector
        :rtype: [float]
        :raises TypeError: If `amount` is not an int
        :raises ValueError: If `amount` is negative
        """
        if type(amount) is not int:
            raise TypeError('amount needs to be a non-negative integer')
        if amount < 0:
            raise ValueError('amount needs to be a non-negative integer')
        if self.order - amount < 2:
            raise ValueError('cannot lower order to less than linears')

        p = self.order - amount
        knots = [ [k] * max(p-1-self.continuity(k), 1) for k in self.knot_spans(True)]
        knots = [ k for sublist in knots for k in sublist]

        if self.periodic > -1:
            # remove excessive ghost knots which appear at both ends of the knot vector
            n0 =                   bisect_left(knot_spans, self.start())
            n1 = len(knot_spans) - bisect_left(knot_spans, self.end())   - 1
            knots = knots[n0*amount : -n1*amount]

        return BSplineBasis(p, knots, self.periodic)

    def insert_knot(self, new_knot):
        """Inserts a knot in the knot vector.

        The return value is a sparse matrix *C* (actually, a dense matrix with
        lots of zeros), such that *N_new* = *N_old* x *C*, where *N* are row
        vectors of basis functions.

        :param float new_knot: The parametric coordinate of the point to insert
        :return: Transformation matrix *C*
        :rtype: numpy.array
        :raises ValueError: If the new knot is outside the domain
        """
        if self.periodic >= 0:
            if new_knot < self.start() or new_knot > self.end():
                new_knot = (new_knot - self.start()) % (self.end() - self.start()) + self.start()
        elif new_knot < self.start() or self.end() < new_knot:
            raise ValueError('new_knot out of range')
        # mu is the index of last non-zero (old) basis function
        mu = bisect_right(self.knots, new_knot)
        n = self.num_functions()
        p = self.order
        C = np.zeros((n + 1, n))
        # the modulus operator i%n in the C-matrix is needed for periodic basis functions
        for i in range(mu - p):
            C[i % (n + 1), i % n] = 1
        for i in range(mu - p, mu):
            if self.knots[i + p - 1] <= new_knot and new_knot <= self.knots[i + p]:
                C[i % (n + 1), i % n] = 1
            else:
                C[i % (n + 1), i % n] = (new_knot - self.knots[i]) / (
                    self.knots[i + p - 1] - self.knots[i])
            if self.knots[i] <= new_knot and new_knot <= self.knots[i + 1]:
                C[(i + 1) % (n + 1), i % n] = 1
            else:
                C[(i + 1) % (n + 1), i % n] = (self.knots[i + p] - new_knot) / (
                    self.knots[i + p] - self.knots[i + 1])
        for i in range(mu, n + 1):
            C[i % (n + 1), (i - 1) % n] = 1

        self.knots = np.insert(self.knots, mu, new_knot)

        # make sure that it is correct periodic after knot insertion
        if self.periodic > -1:
            m  = len(self.knots)
            r  = self.periodic
            if mu <= p+r: # need to fix ghost knots on right side
                k0 = self.knots[0]
                k1 = self.knots[-p-r-1]
                for i in range(p+r+1):
                    self.knots[m-p-r-1+i] = k1 + (self.knots[i]-k0)
            elif mu >= m-p-r-1: # need to fix ghost knots on left side
                k0 = self.knots[p+r]
                k1 = self.knots[-1]
                for i in range(p+r+1):
                    self.knots[i] = k0 - (k1-self.knots[m-p-r-1+i])
        return C

    def roll(self, new_start):
        """rotate a periodic knot vector by setting a new starting index.

        :param int new_start: The index of to the new first knot
        """
        if self.periodic < 0:
            raise  RuntimeError("roll only applicable for periodic knot vectors")

        p = self.order
        k = self.periodic
        n = len(self.knots)
        t1 = self.knots[0] - self.knots[-p - k - 1]
        left  = slice(new_start, n-p-k-1, None)
        len_left = left.stop - left.start
        right = slice(0, n-len_left, None)
        (self.knots[:len_left], self.knots[len_left:]) = (self.knots[left], self.knots[right] - t1)

    def matches(self, bspline, reverse=False):
        """ Checks if this basis equals another basis, when disregarding
        scaling and translation of the knots vector. I.e. will this basis and
        *bspline* yield the same spline object if paired with identical
        controlpoints """
        if self.order != bspline.order or self.periodic != bspline.periodic:
            return False
        dt  = self.knots[-1]    - self.knots[0]
        dt2 = bspline.knots[-1] - bspline.knots[0]
        if reverse:
            return np.allclose( (self.knots[-1]-self.knots[::-1]) / dt,
                                (bspline.knots-bspline.knots[0]) / dt2,
                                atol=state.knot_tolerance)
        else:
            return np.allclose( (self.knots-self.knots[0]) / dt,
                                (bspline.knots-bspline.knots[0]) / dt2,
                                atol=state.knot_tolerance)

    def snap(self, t):
        """  Snap evaluation points to knots if they are sufficiently close
        as given in by state.state.knot_tolerance. This will modify the input vector t

        :param t: evaluation points
        :type  t: [float]
        :return: none
        """
        n = len(self.knots)
        for j in range(len(t)):
            i = bisect_left(self.knots, t[j])
            if i < n and abs(self.knots[i]-t[j]) < state.knot_tolerance:
                t[j] = self.knots[i]
            elif i > 0 and abs(self.knots[i-1]-t[j]) < state.knot_tolerance:
                t[j] = self.knots[i-1]


    def clone(self):
        """Clone the object."""
        return copy.deepcopy(self)

    __call__ = evaluate

    def __len__(self):
        """Returns the number of knots in this basis."""
        return len(self.knots)

    def __getitem__(self, i):
        """Returns the knot at a given index."""
        return self.knots[i]

    def __iadd__(self, a):
        self.knots += a
        return self

    def __isub__(self, a):
        self.knots -= a
        return self

    def __imul__(self, a):
        self.knots *= a
        return self

    def __itruediv__(self, a):
        self.knots /= a
        return self

    __ifloordiv__ = __itruediv__  # integer division (should not distinguish)
    __idiv__ = __itruediv__  # python2 compatibility

    def __repr__(self):
        result = 'p=' + str(self.order) + ', ' + str(self.knots)
        if self.periodic > -1:
            result += ', C' + str(self.periodic) + '-periodic'
        return result



from itertools import product, chain
import re
import warnings

import numpy as np
from tqdm import tqdm
import cv2
import h5py
from scipy.spatial import Delaunay
from scipy.spatial.qhull import QhullError

from ..surface import Surface
from ..volume import Volume
from ..splineobject import SplineObject
from ..basis import BSplineBasis
from ..utils import ensure_listlike
from .. import surface_factory, curve_factory, volume_factory

from .master import MasterIO
from .g2 import G2


class Box(object):

    def __init__(self, x):
        self.x = x


class DiscontBoxMesh(object):

    def __init__(self, n, coord, zcorn):
        nx, ny, nz = n

        X  = np.empty(n + 1, dtype=object)
        Xz = np.zeros((nx + 1, ny + 1, 2 * nz, 3))
        cells = np.empty(n, dtype=object)

        for i, j, k in product(range(nx), range(ny), range(nz)):
            x = []
            for k0, j0, i0 in product(range(2), repeat=3):
                # Interpolate to find the x,y values of this point
                zmin, zmax = coord[i+i0, j+j0, :, 2]
                z  = zcorn[2*i+i0, 2*j+j0, 2*k+k0]
                t = (z - zmax) / (zmin - zmax)
                point = coord[i+i0, j+j0, 0] * t + coord[i+i0, j+j0, 1] * (1 - t)
                x.append(point)

                if X[i+i0,j+j0,k+k0] is None:
                    X[i+i0,j+j0,k+k0] = [point]
                else:
                    X[i+i0,j+j0,k+k0].append(point)
                Xz[i+i0,j+j0,2*k+k0,:] = point

            cells[i,j,k] = Box(x)

        self.X = X
        self.Xz = Xz
        self.n = n

        def hull_or_none(x):
            try:
                return Delaunay(x)
            except QhullError:
                return None

        self.plane_hull = np.array([
            [Delaunay(np.reshape(coord[i:i+2, j:j+2, :, :], (8,3))) for j in range(ny)]
            for i in range(nx)
        ], dtype=object)

        self.hull = np.array([
            [[hull_or_none(cell.x) for cell in cell_tower] for cell_tower in cells_tmp]
            for cells_tmp in cells
        ], dtype=object)

    def cell_at(self, x, guess=None):
        # First, find the 'tower' containing x
        check = -1
        last_i = last_j = 0
        numb_hits = []
        if guess is not None:
            i, j, _ = guess
            check = self.plane_hull[i,j].find_simplex(x)
            # if check > -1: print('correct tower!')
            if check >= 0:
                numb_hits += [(i,j)]
                last_i = i
                last_j = j
            check = -1
        if check == -1:
            for (i, j), hull in np.ndenumerate(self.plane_hull):
                check = hull.find_simplex(x)
                if check >= 0:
                    numb_hits += [(i,j)]
                    last_i = i
                    last_j = j
        i,j = last_i,last_j

#        if len(numb_hits) != 1:
#            print(numb_hits)
#            print(x)
#            print(guess)
#            print(check)

        # assert check >= 0
        assert len(numb_hits) >= 1

        # Find the correct cell in the 'tower'
        check = -1
        if guess is not None:
            _, _, k = guess
            check = self.hull[i,j,k].find_simplex(x)
            # if check > -1: print('correct cell!')
        if check == -1:
            for (i,j) in numb_hits:
                for k, hull in enumerate(self.hull[i,j,:]):
                    if hull is None: continue
                    check = hull.find_simplex(x)
                    if check >= 0: break
                if check >= 0: break

        if check < 0:
            print(numb_hits)
            print(x)
            print(guess)
            print(check)

        # print(f'Returns {i} {j} {k} : {check}')
        assert check >= 0
        return i, j, k

    def get_c0_avg(self):
        """Compute best-approximation vertices for a continuous mesh by averaging the location of all
        corners that 'should' coincide.
        """
        return np.array([[[np.mean(k,axis=0) for k in j] for j in i] for i in self.X])

    def get_discontinuous_all(self):
        """Return a list of vertices suitable for a fully discontinuous mesh."""
        return list(chain.from_iterable(xs[::-1] for xs in self.X.T.flat))

    def get_discontinuous_z(self):
        """Return a list of vertices suitable for a mixed continuity mesh."""
        return self.Xz


class GRDECL(MasterIO):

    def __init__(self, filename):
        if not filename.endswith('.grdecl'):
            filename += '.grdecl'
        self.filename = filename
        self.attribute = {}

    def __enter__(self):
        self.fstream    = open(self.filename, 'r')
        self.line_number = 0
        return self

    def read_specgrid(self):
        args = next(self.fstream).strip().split()
        return np.array(args[:3], dtype=np.int32)

    def read_coord(self):
        nx, ny = self.n[:2]
        ans = np.zeros((nx + 1, ny + 1, 2, 3))
        for j, i in product(range(ny+1), range(nx+1)):
            args = next(self.fstream).split()
            ans[i,j,0,:] = np.array(args[:3], dtype=np.float64)
            ans[i,j,1,:] = np.array(args[3:], dtype=np.float64)
        return ans

    def read_zcorn(self):
        ntot = np.prod(self.n)*8
        numbers = []
        while len(numbers) < ntot:
            numbers += next(self.fstream).split()
        numbers = numbers[:ntot] # strip away any '/' characters at the end of the line
        return np.reshape(np.array(numbers, dtype=np.float64), self.n*2, order='F')

    def cell_property(self, dtype=np.float64):
        ntot = np.prod(self.n)
        numbers = []
        while len(numbers) < ntot:
            numbers += next(self.fstream).split()
        numbers = numbers[:ntot] # strip away any '/' characters at the end of the line
        return np.array(numbers, dtype=dtype)

    def read(self):
        for line in self.fstream:
            line = line.strip().lower()
            if line == 'specgrid':
                self.n = self.read_specgrid()
            elif line == 'coord':
                self.coord = self.read_coord()
            elif line == 'zcorn':
                self.zcorn = self.read_zcorn()
            elif line in {'actnum', 'permx', 'permy', 'permz', 'poro', 'satnum', 'rho', 'kx', 'kz', 'emodulus25', 'poissonratio25', 'pressure', }:
                dtype = np.int32 if line in {'actnum', 'satnum'} else np.float64
                self.attribute[line] = self.cell_property(dtype)
            elif line in {'grid', '/', ''} or line.startswith('--'):
                pass
            elif not re.match('[0-9]', line[0]):
                warnings.showwarning(
                    'Unkown keyword "{}" encountered in file'.format(line.split()[0]),
                    SyntaxWarning, self.filename, self.line_number, line=[],
                )
            else:
                pass # silently skip large number blocks

        self.raw = DiscontBoxMesh(self.n, self.coord, self.zcorn)


    def get_c0_mesh(self):
        # Create the C0-mesh
        nx, ny, nz = self.n
        X = self.raw.get_c0_avg()
        b1 = BSplineBasis(2, [0] + [i/nx for i in range(nx+1)] + [1])
        b2 = BSplineBasis(2, [0] + [i/ny for i in range(ny+1)] + [1])
        b3 = BSplineBasis(2, [0] + [i/nz for i in range(nz+1)] + [1])
        c0_vol = volume_factory.interpolate(X, [b1, b2, b3])
        return c0_vol

    def get_cm1_mesh(self):
        # Create the C^{-1} mesh
        nx, ny, nz = self.n
        Xm1 = self.raw.get_discontinuous_all()
        b1 = BSplineBasis(2, sorted(list(range(self.n[0]+1))*2))
        b2 = BSplineBasis(2, sorted(list(range(self.n[1]+1))*2))
        b3 = BSplineBasis(2, sorted(list(range(self.n[2]+1))*2))
        discont_vol = Volume(b1, b2, b3, Xm1)
        return discont_vol

    def get_mixed_cont_mesh(self):
        # Create mixed discontinuity mesh: C^0, C^0, C^{-1}
        nx, ny, nz = self.n
        Xz = self.raw.get_discontinuous_z()
        b1 = BSplineBasis(2, sorted(list(range(self.n[0]+1))+[0,self.n[0]]))
        b2 = BSplineBasis(2, sorted(list(range(self.n[1]+1))+[0,self.n[1]]))
        b3 = BSplineBasis(2, sorted(list(range(self.n[2]+1))*2))
        true_vol = Volume(b1, b2, b3, Xz, raw=True)
        return true_vol

    def texture(self, p, ngeom, ntexture, method='full', irange=[None,None], jrange=[None,None]):
        # Set the dimensions of geometry and texture map
        # ngeom    = np.floor(self.n / (p-1))
        # ntexture = np.floor(self.n * n)
        # ngeom    = ngeom.astype(np.int32)
        # ntexture = ntexture.astype(np.int32)
        ngeom    = ensure_listlike(ngeom, 3)
        ntexture = ensure_listlike(ntexture, 3)
        p        = ensure_listlike(p, 3)


        # Create the geometry
        ngx, ngy, ngz = ngeom
        b1 = BSplineBasis(p[0], [0]*(p[0]-1) + [i/ngx for i in range(ngx+1)] + [1]*(p[0]-1))
        b2 = BSplineBasis(p[1], [0]*(p[1]-1) + [i/ngy for i in range(ngy+1)] + [1]*(p[1]-1))
        b3 = BSplineBasis(p[2], [0]*(p[2]-1) + [i/ngz for i in range(ngz+1)] + [1]*(p[2]-1))

        l2_fit = surface_factory.least_square_fit
        vol = self.get_c0_mesh()

        i = slice(irange[0], irange[1], None)
        j = slice(jrange[0], jrange[1], None)
        # special case number of evaluation points for full domain
        if irange[1] == None: irange[1] = vol.shape[0]
        if jrange[1] == None: jrange[1] = vol.shape[1]
        if irange[0] == None: irange[0] = 0
        if jrange[0] == None: jrange[0] = 0

        nu = np.diff(irange)
        nv = np.diff(jrange)
        nw = vol.shape[2]

        u = np.linspace(0, 1, nu)
        v = np.linspace(0, 1, nv)
        w = np.linspace(0, 1, nw)

        crvs = []
        crvs.append(curve_factory.polygon(vol[i          ,jrange[0]  , 0,:].squeeze()))
        crvs.append(curve_factory.polygon(vol[i          ,jrange[0]  ,-1,:].squeeze()))
        crvs.append(curve_factory.polygon(vol[i          ,jrange[1]-1, 0,:].squeeze()))
        crvs.append(curve_factory.polygon(vol[i          ,jrange[1]-1,-1,:].squeeze()))
        crvs.append(curve_factory.polygon(vol[irange[0]  ,j          , 0,:].squeeze()))
        crvs.append(curve_factory.polygon(vol[irange[0]  ,j          ,-1,:].squeeze()))
        crvs.append(curve_factory.polygon(vol[irange[1]-1,j          , 0,:].squeeze()))
        crvs.append(curve_factory.polygon(vol[irange[1]-1,j          ,-1,:].squeeze()))
        crvs.append(curve_factory.polygon(vol[irange[0]  ,jrange[0]  , :,:].squeeze()))
        crvs.append(curve_factory.polygon(vol[irange[0]  ,jrange[1]-1, :,:].squeeze()))
        crvs.append(curve_factory.polygon(vol[irange[1]-1,jrange[0]  , :,:].squeeze()))
        crvs.append(curve_factory.polygon(vol[irange[1]-1,jrange[1]-1, :,:].squeeze()))
#        with G2('curves.g2') as myfile:
#            myfile.write(crvs)
#        print('Written curve.g2')


        if method == 'full':
            bottom = l2_fit(vol[i,          j,          0,:].squeeze(), [b1, b2], [u, v])
            top    = l2_fit(vol[i,          j,         -1,:].squeeze(), [b1, b2], [u, v])
            left   = l2_fit(vol[irange[0]  ,j,          :,:].squeeze(), [b2, b3], [v, w])
            right  = l2_fit(vol[irange[1]-1,j,          :,:].squeeze(), [b2, b3], [v, w])
            front  = l2_fit(vol[i,          jrange[0],  :,:].squeeze(), [b1, b3], [u, w])
            back   = l2_fit(vol[i,          jrange[1]-1,:,:].squeeze(), [b1, b3], [u, w])
            volume = volume_factory.edge_surfaces([left, right, front, back, bottom, top])

        elif method == 'z':
            bottom = l2_fit(vol[i,j, 0,:].squeeze(), [b1, b2], [u, v])
            top    = l2_fit(vol[i,j,-1,:].squeeze(), [b1, b2], [u, v])
            volume = volume_factory.edge_surfaces([bottom, top])
            volume.set_order(*p)
            volume.refine(ngz - 1, direction='w')

        volume.reverse(direction=2)

        # Point-to-cell mapping
        # TODO: Optimize more
        eps = 1e-2
        u = [np.linspace(eps, 1-eps, n) for n in ntexture]
        points = volume(*u).reshape(-1, 3)
        cellids = np.zeros(points.shape[:-1], dtype=int)
        cell = None
        nx, ny, nz = self.n
        for ptid, point in enumerate(tqdm(points, desc='Inverse mapping')):
            i, j, k = cell = self.raw.cell_at(point) # , guess=cell)
            cellid = i*ny*nz + j*nz + k
            cellids[ptid] = cellid

        cellids = cellids.reshape(tuple(ntexture))

        all_textures = {}
        for name in self.attribute:
            data = self.attribute[name][cellids]

            # TODO: This flattens the image if it happens to be 3D (or higher...)
            # do we need a way to communicate the structure back to the caller?
            # data = data.reshape(-1, data.shape[-1])

            # TODO: This normalizes the image,
            # but we need a way to communicate the ranges back to the caller
            # a, b = min(data.flat), max(data.flat)
            # data = ((data - a) / (b - a) * 255).astype(np.uint8)

            all_textures[name] = data
        all_textures['cellids'] = cellids

        return volume, all_textures

    def to_ifem(self, p, ngeom, ntexture, method='full', irange=[None,None], jrange=[None,None]):
        translate = {
            'emodulus25'    : 'stiffness',
            'kx'            : 'permx',
            'ky'            : 'permy',
            'kz'            : 'permz',
            'poissonratio25': 'poisson'}

        h5_filename = 'textures.h5'
        h5_file = h5py.File(h5_filename, 'w')
        vol, textures = self.texture(p, ngeom, ntexture, method, irange, jrange)

        # augment dataset with missing information
        if 'kx' in textures and not 'ky' in textures:
            textures['ky'] = textures['kx']

        # print information to png-images and hdf5-files
        print(r'<porotexturematerial>')
        for name, data in textures.items():
            # translate to more IFEM-friendly terminology
            if name in translate: name = translate[name]

            h5_file.create_dataset(name, data=data, compression='gzip')
            a, b = min(data.flat), max(data.flat)
            img  = ((data - a) / (b - a) * 255).astype(np.uint8)
            n    = data.shape
            img  = img.reshape(n[0], n[1]*n[2])
            print('  <property file="{}.png" min="{}" max="{}" name="{}" nx="{}" ny="{}" nz="{}"/>'.format(name, a,b, name, n[0], n[1], n[2]))
            cv2.imwrite(name+'.png', img)
        print(r'</porotexturematerial>')
        h5_file.close()
        print('Written {}'.format(h5_filename))

        with G2('geom.g2') as myfile:
            myfile.write(vol)

    def __exit__(self, exc_type, exc_value, traceback):
        self.fstream.close()



__doc__ = 'Implementation of various refinement schemes.'

from math import atan, pi, tan

import numpy as np

from . import ensure_listlike, check_direction

# TODO: put control over these tolerances somewhere. Modstate in splipy seems
#       to be the place for it, but we can't let splipy.utils influence the
#       structure of splipy.
def knot_exists(existing_knots, new_knot):
    return np.any(np.isclose(existing_knots, new_knot, atol=1e-7, rtol=1e-10))


def geometric_refine(obj, alpha, n, direction=0, reverse=False):
    """geometric_refine(obj, alpha, n, [direction=0], [reverse=False])

    Refine a spline object by making a geometric distribution of element sizes.

    :param obj: The object to refine
    :type obj: :class:`splipy.SplineObject`
    :param float alpha: The length ratio between two sequential knot segments
    :param int n: The number of knots to insert
    :param direction: The direction to refine in
    :param bool reverse: Set to `True` to refine towards the other end
    """
    # some error tests on input
    if n<=0:
        raise ValueError('n should be greater than 0')

    direction = check_direction(direction, obj.pardim)
    if reverse:
        obj.reverse(direction)

    # fetch knots
    knots = obj.knots()
    knot_start = knots[direction][0]
    knot_end   = knots[direction][-1]
    dk = knot_end - knot_start

    # evaluate the factors
    n = n+1 # redefine n to be knot spans instead of new internal knots
    totProd = 1.0
    totSum  = 0.0
    for i in range(n):
        totSum  += totProd
        totProd *= alpha
    d1 = 1.0 / totSum
    knot = d1

    # compute knot locations
    new_knots = []
    for i in range(n-1):
        k = knot_start + knot*dk
        if not knot_exists(knots[direction], k):
            new_knots.append(k)
        knot += alpha*d1
        d1   *= alpha

    # do the actual knot insertion
    obj.insert_knot(new_knots, direction)

    if reverse:
        obj.reverse(direction)
    return obj

def center_refine(obj, S, n, direction=0):
    """center_refine(obj, S, n, [direction=0])

    Refine an object towards the center in a direction, by sampling an
    tan function.

    :param obj: The object to refine
    :type obj: :class:`splipy.SplineObject`
    :param float S: The slope of the tan function (range 0,pi/2)
    :param int n: The number of knots to insert
    :param direction: The direction to refine in
    """
    # some error tests on input
    if n<=0:
        raise ValueError('n should be greater than 0')
    assert 0 < S < np.pi/2

    direction = check_direction(direction, obj.pardim)

    # fetch knots
    knots = obj.knots()
    knot_start = knots[direction][0]
    knot_end   = knots[direction][-1]
    dk = knot_end - knot_start

    # compute knot locations
    new_knots = []
    max_tan  = tan(S)
    for i in range(1,n+1):
        xi  = -1.0 + 2.0*i/(n+1)
        xi *= S
        k   = knot_start + (tan(xi)+max_tan)/2/max_tan*dk
        if not knot_exists(knots[direction], k):
            new_knots.append(k)

    # do the actual knot insertion
    obj.insert_knot(new_knots, direction)
    return obj


def edge_refine(obj, S, n, direction=0):
    """edge_refine(obj, S, n, [direction=0])

    Refine an object towards both edges in a direction, by sampling an
    arctan function.

    :param obj: The object to refine
    :type obj: :class:`splipy.SplineObject`
    :param float S: The slope of the arctan function.
    :param int n: The number of knots to insert
    :param direction: The direction to refine in
    """
    # some error tests on input
    if n<=0:
        raise ValueError('n should be greater than 0')

    direction = check_direction(direction, obj.pardim)

    # fetch knots
    knots = obj.knots()
    knot_start = knots[direction][0]
    knot_end   = knots[direction][-1]
    dk = knot_end - knot_start

    # compute knot locations
    new_knots = []
    max_atan  = atan(S)
    for i in range(1,n+1):
        xi  = -1.0 + 2.0*i/(n+1)
        xi *= S
        k   = knot_start + (atan(xi)+max_atan)/2/max_atan*dk
        if not knot_exists(knots[direction], k):
            new_knots.append(k)

    # do the actual knot insertion
    obj.insert_knot(new_knots, direction)
    return obj


def _splitvector(len, parts):
    delta = len // parts
    sizes = [delta for i in range(parts)]
    remainder = len-parts*delta
    for i in range(parts-remainder+1, parts):
        sizes[i] = sizes[i]+1
    result = [0]
    for i in range(1,parts):
        result.append(sizes[i]+result[i-1])
    return result


def subdivide(objs, n):
    """Subdivide a list of objects by splitting them up along existing knot
    lines. The resulting partition will roughly the same number of elements on
    all pieces. By splitting along *n* lines, we generate *n* + 1 new blocks.

    The number of subdivisions can be a list of integers: one for each
    direction, or a single integer for uniform sudivision.

    :param objs: Objects to split
    :param n: Number of subdivisions to perform
    :type n: int or [int]
    :return: New objects
    :rtype: [:class:`splipy.SplineObject`]
    """
    pardim = objs[0].pardim # 1 for curves, 2 for surfaces, 3 for volumes
    n = ensure_listlike(n, pardim)

    result = objs
    for d in range(pardim):
        # split all objects so far along direction d
        new_results = []
        for obj in result:
            splitting_points = [obj.knots(d)[i] for i in _splitvector(len(obj.knots(d)), n[d]+1)]
            new_results += obj.split(splitting_points[1:], d)

        # only keep the smallest pieces in our result list
        result = new_results

    return result




# -*- coding: utf-8 -*-

from itertools import combinations, product
from math import atan2, sqrt
import numpy as np

try:
    from collections.abc import Sized
except ImportError:
    from collections import Sized

def is_right_hand(patch, tol=1e-3):
    param = tuple((a+b)/2 for a,b in zip(patch.start(), patch.end()))

    if patch.dimension == patch.pardim == 3:
        du = patch.derivative(*param, d=(1,0,0))
        dv = patch.derivative(*param, d=(0,1,0))
        dw = patch.derivative(*param, d=(0,0,1))

        # Normalize
        du = du / np.linalg.norm(du)
        dv = dv / np.linalg.norm(dv)
        dw = dw / np.linalg.norm(dw)

        # Compare cross product
        return np.dot(dw, np.cross(du, dv)) >= tol

    if patch.dimension == patch.pardim == 2:
        du = patch.derivative(*param, d=(1,0))
        dv = patch.derivative(*param, d=(0,1))

        # Normalize
        du = du / np.linalg.norm(du)
        dv = dv / np.linalg.norm(dv)

        return np.cross(du, dv) >= tol

    raise ValueError("Right-handedness only defined for 2D or 3D patches in 2D or 3D space, respectively")

def rotation_matrix(theta, axis):
    axis = axis / np.sqrt(np.dot(axis, axis))
    a = np.cos(theta / 2)
    b, c, d = -axis*np.sin(theta / 2)
    return np.array([[a*a+b*b-c*c-d*d, 2*(b*c-a*d),     2*(b*d+a*c)],
                      [2*(b*c+a*d),     a*a+c*c-b*b-d*d, 2*(c*d-a*b)],
                      [2*(b*d-a*c),     2*(c*d+a*b),     a*a+d*d-b*b-c*c]])

def sections(src_dim, tgt_dim):
    """Generate all boundary sections from a source dimension to a target
    dimension. For example, `sections(3,1)` generates all edges on a volume.

    The return values are lists of length `src_dim` with each element either 0,
    --1 or `None`, which are suitable for passing to
    :func:`splipy.SplineObject.section`.
    """
    # Enumerate all combinations of fixed directions
    nfixed = src_dim - tgt_dim
    for fixed in combinations(range(src_dim), r=nfixed):
        # Enumerate all {0,-1}^n over the fixed directions
        for indices in product([0, -1], repeat=nfixed):
            args = [None] * src_dim
            for f, i in zip(fixed, indices[::-1]):
                args[f] = i
            yield args

def section_from_index(src_dim, tgt_dim, i):
    """Return the i'th section from a source dimension to a target dimension.

    See :func:`splipy.Utils.sections` for more information.
    """
    for j, s in enumerate(sections(src_dim, tgt_dim)):
        if i == j:
            return s

def section_to_index(section):
    """Return the index corresponding to a section."""
    src_dim = len(section)
    tgt_dim = sum(1 for s in section if s is None)
    for i, t in enumerate(sections(src_dim, tgt_dim)):
        if tuple(section) == tuple(t):
            return i

def check_section(*args, **kwargs):
    """check_section(u, v, ...)

    Parse arguments and return a section spec.

    The keyword argument `pardim` *must* be provided. The return value is a
    section as described in :func:`splipy.Utils.sections`.
    """
    pardim = kwargs['pardim']
    args = list(args)
    while len(args) < pardim:
        args.append(None)
    for k in set(kwargs.keys()) & set('uvw'):
        index = 'uvw'.index(k)
        args[index] = kwargs[k]
    return args

def check_direction(direction, pardim):
    if   direction in {0, 'u', 'U'} and 0 < pardim:
        return 0
    elif direction in {1, 'v', 'V'} and 1 < pardim:
        return 1
    elif direction in {2, 'w', 'W'} and 2 < pardim:
        return 2
    raise ValueError('Invalid direction')

def ensure_flatlist(x):
    """Flattens a multi-list x to a single index list."""
    if isinstance(x[0], Sized):
        return x[0]
    return x

def is_singleton(x):
    """Checks if x is list-like."""
    return not isinstance(x, Sized)

def ensure_listlike(x, dups=1):
    """Wraps x in a list if it's not list-like."""
    try:
        while len(x) < dups:
            x = list(x)
            x.append(x[-1])
        return x
    except TypeError:
        return [x] * dups
    except IndexError:
        return []

def rotate_local_x_axis(xaxis=(1,0,0), normal=(0,0,1)):
    # rotate xaxis vector back to reference domain (r=1, around origin)
    theta = atan2(normal[1], normal[0])
    phi   = atan2(sqrt(normal[0]**2+normal[1]**2), normal[2])
    R1 = rotation_matrix(-theta, (0,0,1))
    R2 = rotation_matrix(-phi,   (0,1,0))
    if len(xaxis) != 3: # typically 2D geometries
        xaxis = [xaxis[0], xaxis[1], 0]
    xaxis = np.array([xaxis])
    xaxis = xaxis.dot(R1).dot(R2)
    # if xaxis is orthogonal to normal, then xaxis[2]==0 now. If not then
    # treating it as such is the closest projection, which makes perfect sense
    return atan2(xaxis[0,1], xaxis[0,0])

def flip_and_move_plane_geometry(obj, center=(0,0,0), normal=(0,0,1)):
    """re-orients a planar geometry by moving it to a different location and
    tilting it"""
    # don't touch it if not needed. translate or scale operations may force
    # object into 3D space
    if not np.allclose(normal, np.array([0,0,1])):
        theta = atan2(normal[1], normal[0])
        phi   = atan2(sqrt(normal[0]**2+normal[1]**2), normal[2])
        obj.rotate(phi,   (0,1,0))
        obj.rotate(theta, (0,0,1))
    if not np.allclose(center, 0):
        obj.translate(center)
    return obj

def reshape(cps, newshape, order='C', ncomps=None):
    """Like numpy's reshape, but preserves control points of several dimensions
    that are stored contiguously.

    The return value has shape (*newshape, ncomps), where ncomps is the number
    of components per control point, as inferred by the size of `cps` and the
    desired shape.

    The `order` argument ('C' or 'F') determines the order in which control
    points are read, but does *not* affect the order in which each component of
    a control point is read.
    """
    npts = np.prod(newshape)
    if ncomps is None:
        try:
            ncomps = cps.size // npts
        except AttributeError:
            ncomps = len(cps) // npts

    if order == 'C':
        shape = list(newshape) + [ncomps]
    elif order == 'F':
        shape = list(newshape[::-1]) + [ncomps]
    cps = np.reshape(cps, shape)
    if order == 'F':
        spec = list(range(len(newshape)))[::-1] + [len(newshape)]
        cps = cps.transpose(spec)
    return cps

def uniquify(iterator):
    """Iterates over all elements in `iterator`, removing duplicates."""
    seen = set()
    for i in iterator:
        if i in seen:
            continue
        seen.add(i)
        yield i

def raise_order_1D(n, k, T, P, m, periodic):
    """ Implementation of method in "Efficient Degree Elevation and Knot Insertion
        for B-spline Curves using Derivatives" by Qi-Xing Huang a Shi-Min Hu, Ralph R Martin. Only the case of open knot vector is fully implemented
    :param int n: (n+1) is the number of initial basis functions
    :param int k: spline order
    :param T: knot vector
    :param P: weighted NURBS coefficients
    :param int m: number of degree elevations
    :param int periodic: Number of continuous derivatives at start and end. -1 is not periodic, 0 is continuous, etc.
    :return Q: new control points
    """

    u = np.unique(T[k-1:-k+1])
    S = u.size - 1
    d = P.shape[0] # dimension of spline

    # Find multiplicity of the knot vector T
    b = BSplineBasis(k, T)
    z = [k-1-b.continuity(t0) for t0 in b.knot_spans()]
    
    # Step 1: Find Pt_i^j
    Pt = np.zeros((d,n+1,k))
    Pt[:,:,0] = P
    Pt = np.concatenate((Pt,Pt[:,0:periodic+1,:]),axis=1) 
    n += periodic+1

    beta = np.cumsum(z[1:-1],dtype=int)
    beta = np.insert(beta,0,0) # include the empty sum (=0)
    for l in range(1,k):
        for i in range(0,n+1-l):
            if T[i+l] < T[i+k]:
                Pt[:,i,l] = (Pt[:,i+1,l-1] - Pt[:,i,l-1])/(T[i+k]-T[i+l])

    # Step 2: Create new knot vector Tb 
    nb = n + S*m
    Tb = np.zeros(nb+m+k+1)
    Tb[:k-1] = T[:k-1]
    Tb[-k+1:] = T[-k+1:]
    j = k-1 
    for i in range(0,len(z)):
        Tb[j:j+z[i]+m] = u[i]
        j = j+z[i]+m
    
    # Step 3: Find boundary values of Qt_i^j
    Qt = np.zeros((d,nb+1,k))
    l_arr = np.array(range(1,k))
    alpha = np.cumprod((k-l_arr)/(k+m-l_arr))
    alpha = np.insert(alpha,0,1) # include the empty product (=1)
    indices = range(0,k)
    Qt[:,0,indices] = np.multiply(Pt[:,0,indices],np.reshape(alpha[indices],(1,1,k))) # (21)
    for p in range(0,S):
        indices = range(k-z[p],k)
        Qt[:,beta[p]+p*m,indices] = np.multiply(Pt[:,beta[p],indices],np.reshape(alpha[indices],(1,1,z[p]))) # (22)
    
    for p in range(0,S):
        idx = beta[p]+p*m
        Qt[:,idx+1:m+idx+1,k-1] = np.repeat(Qt[:,idx:idx+1,k-1],m,axis=1) # (23)
    
    # Step 4: Find remaining values of Qt_i^j
    for j in range(k-1,0,-1):
        for i in range(0,nb):
            if Tb[i+k+m] > Tb[i+j]:
                Qt[:,i+1,j-1] = Qt[:,i,j-1] + (Tb[i+k+m] - Tb[i+j])*Qt[:,i,j] #(20) with Qt replacing Pt

    return Qt[:,:,0]

__all__ = [
    'nutils', 'refinement', 'image', 'NACA', 'curve', 'smooth',
    'rotation_matrix', 'sections', 'section_from_index', 'section_to_index',
    'check_section', 'check_direction', 'ensure_flatlist', 'is_singleton',
    'ensure_listlike', 'rotate_local_x_axis', 'flip_and_move_plane_geometry',
    'reshape','raise_order_1D'
]



from itertools import islice

import numpy as np

from ..curve import Curve
from ..surface import Surface
from ..volume import Volume
from ..splineobject import SplineObject
from ..basis import BSplineBasis

from .master import MasterIO


class SPL(MasterIO):

    def __init__(self, filename):
        if not filename.endswith('.spl'):
            filename += '.spl'
        self.filename = filename
        self.trimming_curves = []

    def __enter__(self):
        self.fstream = open(self.filename, 'r')
        return self

    def lines(self):
        for line in self.fstream:
            yield line.split('#', maxsplit=1)[0].strip()

    def read(self):
        lines = self.lines()

        version = next(lines).split()
        assert version[0] == 'C'
        assert version[3] == '0' # No support for rational SPL yet
        pardim = int(version[1])
        physdim = int(version[2])

        orders = [int(k) for k in islice(lines, pardim)]
        ncoeffs = [int(k) for k in islice(lines, pardim)]
        totcoeffs = int(np.prod(ncoeffs))
        nknots = [a + b for a, b in zip(orders, ncoeffs)]

        next(lines) # Skip spline accuracy

        knots = [[float(k) for k in islice(lines, nkts)] for nkts in nknots]
        bases = [BSplineBasis(p, kts, -1) for p, kts in zip(orders, knots)]

        cpts = np.array([float(k) for k in islice(lines, totcoeffs * physdim)])
        cpts = cpts.reshape(physdim, *(ncoeffs[::-1])).transpose()

        if pardim == 1:
            patch = Curve(*bases, controlpoints=cpts, raw=True)
        elif pardim == 2:
            patch = Surface(*bases, controlpoints=cpts, raw=True)
        elif pardim == 3:
            patch = Volume(*bases, controlpoints=cpts, raw=True)
        else:
            patch = SplineObject(bases, controlpoints=cpts, raw=True)

        return [patch]

    def __exit__(self, exc_type, exc_value, traceback):
        self.fstream.close()



#coding:utf-8

import struct

import numpy as np

from ..surface import Surface
from ..volume import Volume
from ..utils import ensure_listlike
from ..splinemodel import SplineModel

from .master import MasterIO


ASCII_FACET = """facet normal 0 0 0
outer loop
vertex {face[0][0]:.4f} {face[0][1]:.4f} {face[0][2]:.4f}
vertex {face[1][0]:.4f} {face[1][1]:.4f} {face[1][2]:.4f}
vertex {face[2][0]:.4f} {face[2][1]:.4f} {face[2][2]:.4f}
endloop
endfacet
"""

BINARY_HEADER ="80sI"
BINARY_FACET = "12fH"


class ASCII_STL_Writer(object):
    """ Export 3D objects build of 3 or 4 vertices as ASCII STL file.
    """
    def __init__(self, stream):
        self.fp = stream
        self._write_header()

    def _write_header(self):
        self.fp.write("solid python\n")

    def close(self):
        self.fp.write("endsolid python\n")

    def _write(self, face):
        self.fp.write(ASCII_FACET.format(face=face))

    def _split(self, face):
        p1, p2, p3, p4 = face
        return (p1, p2, p3), (p3, p4, p1)

    def add_face(self, face):
        """ Add one face with 3 or 4 vertices. """
        if len(face) == 4:
            face1, face2 = self._split(face)
            self._write(face1)
            self._write(face2)
        elif len(face) == 3:
            self._write(face)
        else:
            raise ValueError('only 3 or 4 vertices for each face')

    def add_faces(self, faces):
        """ Add many faces. """
        for face in faces:
            self.add_face(face)

class BINARY_STL_Writer(ASCII_STL_Writer):
    """ Export 3D objects build of 3 or 4 vertices as binary STL file.
    """
    def __init__(self, stream):
        self.counter = 0
        #### new-style classes way of calling super constructor
        # super(Binary_STL_Writer, self).__init__(stream)

        #### old-style classes way of doing it
        ASCII_STL_Writer.__init__(self, stream)

    def close(self):
        self._write_header()

    def _write_header(self):
        self.fp.seek(0)
        self.fp.write(struct.pack(BINARY_HEADER, b'Python Binary STL Writer', self.counter))

    def _write(self, face):
        self.counter += 1
        data = [
            0., 0., 0.,
            face[0][0], face[0][1], face[0][2],
            face[1][0], face[1][1], face[1][2],
            face[2][0], face[2][1], face[2][2],
            0
        ]
        self.fp.write(struct.pack(BINARY_FACET, *data))


class STL(MasterIO):
    def __init__(self, filename, binary=True):
        if filename[-4:] != '.stl':
            filename += '.stl'
        self.filename = filename
        self.binary   = binary

    def __enter__(self):
        if self.binary:
            fp = open(self.filename, 'wb')
            self.writer = BINARY_STL_Writer(fp)
        else:
            fp = open(self.filename, 'w')
            self.writer = ASCII_STL_Writer(fp)
        return self

    def write(self, obj, n=None):
        if isinstance(obj, SplineModel):
            if obj.pardim == 3: # volume model
                for surface in obj.boundary():
                    self.write_surface(surface.obj,n)
            elif obj.pardim == 2: # surface model
                for surface in obj:
                    self.write_surface(surface, n)

        elif isinstance(obj, Volume):
            for surface in obj.faces():
                self.write_surface(surface, n)

        elif isinstance(obj, Surface):
            self.write_surface(obj, n)

        else:
            raise ValueError('Unsopported object for STL format')

    def write_surface(self, surface, n=None):
        # choose evaluation points as one of three cases:
        #   1. specified with input
        #   2. linear splines, only picks knots
        #   3. general splines choose 2*order-1 per knot span
        if n != None:
            n = ensure_listlike(n,2)

        if n != None:
            u = np.linspace(surface.start(0), surface.end(0), n[0])
        elif surface.order(0) == 2:
            u = surface.knots(0)
        else:
            knots = surface.knots(0)
            p = surface.order(0)
            u = [np.linspace(k0,k1, 2*p-3, endpoint=False) for (k0,k1) in zip(knots[:-1], knots[1:])]
            u = [point for element in u for point in element] + knots
            u = np.sort(u)

        if n != None:
            v = np.linspace(surface.start(1), surface.end(1), n[1])
        elif surface.order(1) == 2:
            v = surface.knots(1)
        else:
            knots = surface.knots(1)
            p = surface.order(1)
            v = [np.linspace(k0,k1, 2*p-3, endpoint=False) for (k0,k1) in zip(knots[:-1], knots[1:])]
            v = [point for element in v for point in element] + knots
            v = np.sort(v)

        # perform evaluation and make sure that we have 3 components (in case of 2D geometries)
        x = surface(u,v)
        if x.shape[2] != 3:
            x.resize((x.shape[0],x.shape[1],3))

        # compute tiny quad pieces
        faces = [[x[i,j], x[i,j+1], x[i+1,j+1], x[i+1,j]] for i in range(x.shape[0]-1) for j in range(x.shape[1]-1)]

        self.writer.add_faces(faces)

    def __exit__(self, exc_type, exc_value, traceback):
        self.writer.close()
        self.writer.fp.close()




# -*- coding: utf-8 -*-

__doc__ = 'Implementation of image based mesh generation.'

from math import sqrt
import sys
import warnings

import numpy as np

from ..basis import BSplineBasis
from .. import curve_factory, surface_factory


def get_corners(X, L=50, R=30, D=15):
    """Detects corners of traced outlines using the SAM04 algorithm.

    The outline is assumed to constitute a discrete closed curve where each
    point is included just once. Increasing `D` and `R` will give the same
    number of corners or fewer corners.

    :param numpy.array X: A traced outline forming a discrete closed curve
        (size *n* × 2)
    :param float L: Controls the scale at which corners are measured.
    :param float R: Controls how close corners can appear.
    :param float D: The lower bound for the corner metric. Corner candidates
        with lower metric than this are rejected.
    :return: The indices of X that consitute corner points
    :rtype: numpy.array
    """
    n = len(X)

    # Finds corner candidates
    d = np.zeros(n)
    for i in range(1,n+1):
        if i+L <= n:
            k = i+L
            index = np.array(range(i+1,k))
        else:
            k = i+L-n
            index = np.array(list(range(i+1,n+1)) + list(range(1,k)))

        M = X[k-1,:]-X[i-1,:]

        if M[0] == 0:
            dCand = abs(X[index-1,0]-X[i-1,0])
        else:
            m = float(M[1])/M[0]
            dCand = abs(X[index-1,1]-m*X[index-1,0]+m*X[i-1,0]-X[i-1,1])/sqrt(m**2+1)

        Y = max(dCand)
        I = np.argmax(dCand)
        if Y > d[index[I]-1]:
            d[index[I]-1] = Y

    I = np.where(d > 0)[0]
    # Rejects candidates which do not meet the lower metric bound D.
    index  = d <  D
    index2 = d >= D
    d[index] = 0
    C = np.array(range(n))
    C = C[index2]


    # Rejects corners that are too close to a corner with larger metric.
    l = len(C)
    j = 0
    while j+1 < l:
        if abs(C[j]-C[j+1]) <= R:
            if d[C[j]] > d[C[j+1]]:
                C = np.delete(C, j+1)
            else:
                C = np.delete(C, j)
            l = l-1
        else:
            j = j+1

    if l > 0 and abs(C[0]+n-C[-1]) <=R:
        if d[C[-1]] > d[C[0]]:
            C = C[1:-1]
        else:
            C = C[0:-2]

    # always include end-points in corner list, and never closer than 4 indices
    if 0 not in C:
        C = np.insert(C,0,0)
    if (n-1) not in C:
        C = np.append(C,n-1)
    remove = []
    for i in range(1,len(C)-1):
        if C[i]-C[i-1] < 5:
            remove.append(i)
    if C[-1]-C[-2] < 5 and len(C)-2 not in remove:
        remove.append(len(C)-2)

    remove.reverse()
    for j in remove:
        C = np.delete(C,j)

    return C


def image_curves(filename):
    """Generate B-spline curves corresponding to the edges in a black/white
    mask image.

    :param str filename: Name of image file to read
    :return: All curves generated by tracing the black/white edges
    :rtype: [:class:`splipy.Curve`]
    """
    import cv2

    im = cv2.imread(filename)

    # initialize image holders
    imGrey   = np.zeros((len(im),   len(im[0])),   np.uint8)
    imBlack  = np.zeros((len(im),   len(im[0])),   np.uint8)

    # convert to greyscale image
    cv2.cvtColor(im, cv2.COLOR_RGB2GRAY, imGrey)

    # convert to binary black/white
    cv2.threshold(imGrey, 128, 255, cv2.THRESH_BINARY, imBlack)

    # find contour curves in image
    if cv2.__version__[0] == '3':
        warnings.warn(FutureWarning('openCV v.3 will eventually be discontinued. Please update your version: \"pip install opencv-python --upgrade\"'))
        [_, contours, _] = cv2.findContours(imBlack, cv2.RETR_LIST, cv2.CHAIN_APPROX_NONE)
    else:
        [contours, _]    = cv2.findContours(imBlack, cv2.RETR_LIST, cv2.CHAIN_APPROX_NONE)

    result = []
    for i in range(len(contours)-1):   # for all contours (except the last one which is the edge)
        pts = contours[i][:,0,:]       # I have no idea why there needs to be a 0 here
        for j in range(len(pts)):      # invert y-axis since images are stored the other way around
            pts[j][1] = len(im[0])-pts[j][1]

        corners = get_corners(pts)
        if len(corners)>2:                        # start/stop tagged as corners. If any inner corners, then
            pts     = np.roll(pts, -corners[1],0) # rearrange points so start/stop falls at a natural corner.
            corners = get_corners(pts)            # recompute corners, since previous sem might be smooth

        n = len(pts)
        parpt = list(range(n))
        for i in range(n):
            parpt[i] = float(parpt[i]) / (n-1)

        # the choice of knot vector is a tricky one. We'll go with the following strategy:
        # - cubic, p=3 curve
        # - C^0 at corner points (computed above)
        # - at least one C^2 knot between corner knots
        # - otherwise as uniform as possible
        # - starts at 0, ends at 1
        # - around 1/10 the number of control points wrt points
        # - up to a max of 100(ish) control points for large models

        # start off with a uniform(ish) knot vector
        knot = []
        nStart = min(n//10, 90)
        for i in range(nStart+1):
            knot.append(int(1.0*i*(n-1)/nStart))
        c = corners.tolist()
        knot = sorted(list(set(knot+c))) # unique sorted list

        # make sure there is at least one knot between corners
        newKnot = []
        for i in range(len(c)-1):
            if knot.index(c[i+1])-knot.index(c[i]) == 1:
                newKnot.append((c[i+1]+c[i])/2)
        knot = sorted(knot + newKnot)

        # make sure no two knots are too close (typical corners which do this)
        for i in range(1,len(knot)-1):
            if knot[i] not in c:
                knot[i] = (knot[i-1]+knot[i+1])/2.0

        # make C^0 at corners and C^-1 at endpoints by duplicating knots
        knot = sorted(knot + c + c + [0,n-1]) # both c and knot contains a copy of the endpoints

        # make it span [0,1] instead of [0,n-1]
        for i in range(len(knot)):
            knot[i] /= float(n-1)

        # make it periodic since these are all closed curves
        knot[0]  -= knot[-1] - knot[-5]
        knot[-1] += knot[4]  - knot[1]

        basis = BSplineBasis(4, knot, 0)

        c = curve_factory.least_square_fit(np.array(pts), basis, parpt)
        result.append(c)

    return result

def image_height(filename, N=[30,30], p=[4,4]):
    """Generate a B-spline surface approximation given by the heightmap in a
    grayscale image.

    :param str filename: Name of image file to read
    :param (int,int) N: Number of controlpoints in u-direction
    :param (int,int) p: Polynomial order (degree+1)
    :return: Normalized (all values between 0 and 1) heightmap approximation
    :rtype: :class:`splipy.Surface`
    """

    import cv2

    im = cv2.imread(filename)

    width  = len(im[0])
    height = len(im)

    # initialize image holder
    imGrey = np.zeros((height, width), np.uint8)

    # convert to greyscale image
    cv2.cvtColor(im, cv2.COLOR_RGB2GRAY, imGrey)

    # guess uniform evaluation points and knot vectors
    u = list(range(width))
    v = list(range(height))
    knot1 = [0]*(p[0]-1) + list(range(N[0]-p[0]+2)) + [N[0]-p[0]+1]*(p[0]-1)
    knot2 = [0]*(p[1]-1) + list(range(N[1]-p[1]+2)) + [N[1]-p[1]+1]*(p[1]-1)

    # normalize all values to be in range [0, 1]
    u     = [float(i)/u[-1]     for i in u]
    v     = [float(i)/v[-1]     for i in v]
    knot1 = [float(i)/knot1[-1] for i in knot1]
    knot2 = [float(i)/knot2[-1] for i in knot2]

    # flip and reverse image so coordinate (0,0) is at lower-left corner
    imGrey = imGrey.T  / 255.0
    imGrey = np.flip(imGrey, axis=1)
    x,y    = np.meshgrid(u,v, indexing='ij')
    pts   = np.stack([x,y,imGrey], axis=2)

    basis1 = BSplineBasis(p[0], knot1)
    basis2 = BSplineBasis(p[1], knot2)

    return surface_factory.least_square_fit(pts,[basis1, basis2], [u,v])

def image_convex_surface(filename):
    """Generate a single B-spline surface corresponding to convex black domain
    of a black/white mask image. The algorithm traces the boundary and searches
    for 4 natural corner points. It will then generate 4 boundary curves which
    will be used to create the surface by Coons Patch.

    :param str filename: Name of image file to read
    :return: B-spline surface
    :rtype: :class:`splipy.Surface`
    """
    # generate boundary curve
    crv = image_curves(filename)

    # error test input
    if len(crv) != 1:
        raise RuntimeError('Error: image_convex_surface expects a single closed curve. Multiple curves detected')

    crv = crv[0]

    # parametric value of corner candidates. These are all in the range [0,1] and both 0 and 1 is present
    kinks = crv.get_kinks()

    # generate 4 corners
    if len(kinks) == 2:
        corners = [0, .25, .5, .75]

    elif len(kinks) == 3:
        corners = [0, (0+kinks[1])/2, kinks[1], (1+kinks[1])/2]

    elif len(kinks) == 4:
        if kinks[1]-kinks[0] > kinks[2]-kinks[1] and kinks[1]-kinks[0] > kinks[3]-kinks[2]:
            corners = [0, (kinks[0]+kinks[1])/2] + kinks[1:3]
        elif kinks[2]-kinks[1] > kinks[3]-kinks[2]:
            corners = [0, kinks[1], (kinks[1]+kinks[2])/2], kinks[2]
        else:
            corners = [0] + kinks[1:3] + [(kinks[2]+kinks[3])/2]

    else:
        while len(kinks) > 5:
            max_span   = 0
            max_span_i = 0
            for i in range(1,len(kinks)-1):
                max_span   = max(max_span, kinks[i+1]-kinks[i-1])
                max_span_i = i
            del kinks[max_span_i]
        corners = kinks[0:4]

    return surface_factory.edge_curves(crv.split(corners))
